"""
The Dialog to Export Data from a Chart
"""

from functools import partial

from qtpy.QtCore import Qt, QSize
from qtpy.QtWidgets import (QFormLayout, QCheckBox, QLabel, QLineEdit,
                            QPushButton, QHBoxLayout, QFontDialog)
from pydm import Display
from pydm.utilities.iconfont import IconFont

from .defaults import X_AXIS_LABEL_SEPARATOR


class AxisSettingsDisplay(Display):
    def __init__(self, main_display, parent=None):
        super(AxisSettingsDisplay, self).__init__(parent=parent)
        self.main_layout = QFormLayout()
        self.main_display = main_display

        self.chart = self.main_display.chart

        self.x_axis_lbl = QLabel("x-axis Label")
        self.x_axis_label_line_edt = QLineEdit()
        current_x_label = self.chart.labels["bottom"]
        if current_x_label:
            current_x_label = current_x_label[
                              current_x_label.find(X_AXIS_LABEL_SEPARATOR) +
                              len(X_AXIS_LABEL_SEPARATOR):]
            self.x_axis_label_line_edt.setText(current_x_label)
        self.x_axis_label_line_edt.textChanged.connect(
            partial(self.handle_axis_label_change, "bottom"))

        self.x_axis_font_btn = QPushButton()
        self.x_axis_font_btn.setMaximumHeight(32)
        self.x_axis_font_btn.setMaximumWidth(32)
        self.x_axis_font_btn.setIcon(IconFont().icon("font"))
        self.x_axis_font_btn.clicked.connect(
            partial(self.handle_font_change, "bottom")
        )

        self.x_axis_unit_lbl = QLabel("x-axis Unit")
        self.x_axis_unit_edt = QLineEdit()
        self.x_axis_unit_edt.setMaximumWidth(150)
        self.x_axis_unit_edt.setText(self.chart.units["bottom"])
        self.x_axis_unit_edt.textChanged.connect(
            partial(self.handle_axis_label_change, "bottom", is_unit=True))

        self.y_axis_lbl = QLabel("y-axis Label")
        self.y_axis_label_line_edt = QLineEdit()
        self.y_axis_label_line_edt.setText(self.chart.labels["left"])
        self.y_axis_label_line_edt.textChanged.connect(
            partial(self.handle_axis_label_change, "left"))

        self.y_axis_font_btn = QPushButton()
        self.y_axis_font_btn.setMaximumHeight(32)
        self.y_axis_font_btn.setMaximumWidth(32)
        self.y_axis_font_btn.setIcon(IconFont().icon("font"))
        self.y_axis_font_btn.clicked.connect(
            partial(self.handle_font_change, "left")
        )


        self.y_axis_unit_lbl = QLabel("y-axis Unit")
        self.y_axis_unit_edt = QLineEdit()
        self.y_axis_unit_edt.setMaximumWidth(150)
        self.y_axis_unit_edt.setText(self.chart.units["left"])
        self.y_axis_unit_edt.textChanged.connect(
            partial(self.handle_axis_label_change, "left", is_unit=True))

        self.right_y_axis_label_line_edt = QLineEdit()

        self.display_right_y_axis_chk = QCheckBox("Display the right y-axis")
        self.display_right_y_axis_chk.setChecked(self.chart.getShowRightAxis())
        self.display_right_y_axis_chk.clicked.connect(
            self.handle_right_y_axis_checkbox_changed)
        self.display_right_y_axis_chk.setChecked(self.chart.getShowRightAxis())

        self.right_y_axis_lbl = None
        self.right_y_axis_unit_edt = None
        self.right_y_axis_unit_lbl = None
        self.right_y_axis_unit_edt = None

        self.close_dialog_btn = QPushButton("Close")
        self.close_dialog_btn.clicked.connect(self.handle_close_button_clicked)

        self.setWindowTitle("Axis Settings")
        self.setMinimumSize(500, 300)
        self.setWindowModality(Qt.ApplicationModal)

        self.setup_ui()

    def ui_filepath(self):
        """
        The path to the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def ui_filename(self):
        """
        The name of the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def setup_ui(self):
        x_axis_layout = QHBoxLayout()
        x_axis_layout.addWidget(self.x_axis_label_line_edt)
        x_axis_layout.addWidget(self.x_axis_font_btn)

        y_axis_layout = QHBoxLayout()
        y_axis_layout.addWidget(self.y_axis_label_line_edt)
        y_axis_layout.addWidget(self.y_axis_font_btn)

        # Add widgets to the form layout
        self.main_layout.setSpacing(10)
        self.main_layout.addRow(self.x_axis_lbl, x_axis_layout)
        self.main_layout.addRow(self.x_axis_unit_lbl, self.x_axis_unit_edt)
        self.main_layout.addRow(self.y_axis_lbl, y_axis_layout)
        self.main_layout.addRow(self.y_axis_unit_lbl, self.y_axis_unit_edt)
        self.main_layout.addRow(self.display_right_y_axis_chk, None)
        self.main_layout.addRow(None, self.close_dialog_btn)

        self.display_right_y_axis_chk.setChecked(self.chart.getShowRightAxis())
        if self.chart.getShowRightAxis():
            self.display_right_y_axis_chk.clicked.emit(
                self.chart.getShowRightAxis())

        self.setLayout(self.main_layout)

    def handle_right_y_axis_checkbox_changed(self, is_checked):
        self.chart.setShowRightAxis(is_checked)

        if is_checked:
            right_label = self.chart.labels["right"]
            if not right_label:
                right_label = self.y_axis_label_line_edt.text()
            right_unit = self.chart.units["right"]
            if not right_unit:
                right_unit = self.y_axis_unit_edt.text()

            self.right_y_axis_font_btn = QPushButton()
            self.right_y_axis_font_btn.setMaximumHeight(32)
            self.right_y_axis_font_btn.setMaximumWidth(32)
            self.right_y_axis_font_btn.setIcon(IconFont().icon("font"))
            self.right_y_axis_font_btn.clicked.connect(
                partial(self.handle_font_change, "right")
            )

            self.right_y_axis_lbl = QLabel("Right y-axis Label")
            self.right_y_axis_label_line_edt = QLineEdit()
            self.right_y_axis_label_line_edt.textChanged.connect(
                partial(self.handle_axis_label_change, "right"))

            self.right_y_axis_unit_lbl = QLabel("Right y-axis Unit")
            self.right_y_axis_unit_edt = QLineEdit()
            self.right_y_axis_unit_edt.setMaximumWidth(150)
            self.right_y_axis_unit_edt.textChanged.connect(
                partial(self.handle_axis_label_change, "right",
                        is_unit=True))

            self.right_y_axis_label_line_edt.setText(right_label)
            self.right_y_axis_unit_edt.setText(right_unit)

            right_y_axis_layout = QHBoxLayout()
            right_y_axis_layout.addWidget(self.right_y_axis_label_line_edt)
            right_y_axis_layout.addWidget(self.right_y_axis_font_btn)

            self.main_layout.insertRow(self.main_layout.rowCount() - 1,
                                       self.right_y_axis_lbl,
                                       right_y_axis_layout)
            self.main_layout.insertRow(self.main_layout.rowCount() - 1,
                                       self.right_y_axis_unit_lbl,
                                       self.right_y_axis_unit_edt)
        else:
            self.chart.showAxis("right", show=False)
            self.right_y_axis_lbl.deleteLater()
            self.right_y_axis_label_line_edt.deleteLater()
            self.right_y_axis_unit_lbl.deleteLater()
            self.right_y_axis_unit_edt.deleteLater()
            self.right_y_axis_font_btn.deleteLater()

    def handle_axis_label_change(self, axis_position, new_label, is_unit=False):
        if is_unit:
            self.chart.setLabel(axis_position, units=new_label)
            self.chart.units[axis_position] = new_label
        else:
            if axis_position == "bottom":
                current_label = self.chart.getBottomAxisLabel()
                if X_AXIS_LABEL_SEPARATOR in current_label:
                    current_label = current_label[:current_label.find(
                        X_AXIS_LABEL_SEPARATOR) - len(X_AXIS_LABEL_SEPARATOR)]
                new_label = current_label + X_AXIS_LABEL_SEPARATOR + new_label
            self.chart.setLabel(axis_position, text=new_label)
            self.chart.labels[axis_position] = new_label
        return

    def change_axis_font(self, axis, font):
        axis.setTickFont(font)
        axis.label.setFont(font)

    def handle_font_change(self, axis_position):
        axis = self.chart.getAxis(axis_position)
        label = axis.label
        initial = axis.tickFont

        dialog = QFontDialog(self)
        if initial:
            dialog.setCurrentFont(initial)
        dialog.setOption(QFontDialog.DontUseNativeDialog, True)
        dialog.fontSelected.connect(partial(self.change_axis_font, axis))
        dialog.open()

    def closeEvent(self, event):
        self.handle_close_button_clicked()

    def handle_close_button_clicked(self):
        """
        Close the dialog when the Close button is clicked.
        """
        self.close()



import os
try:
    from os import errno
except ImportError:
    # For Python 3.7
    import errno

import sys
import traceback
import argparse
import logging

import warnings
warnings.filterwarnings("ignore")

from pydm.application import PyDMApplication

from qtpy import QtGui, QtCore

import timechart
from timechart.displays.main_display import TimeChartDisplay


def main():
    args, extra_args = _parse_arguments()

    try:
        os.makedirs("logs")
    except os.error as err:
        # It's OK if the log directory exists. This is to be compatible with Python 2.7
        if err.errno != errno.EEXIST:
            raise err

    logging.basicConfig(level=logging.INFO, filename=os.path.join("logs", "timechart.log"),
                        format="[%(asctime)s] [%(levelname)-8s] - %(message)s")

    logger = logging.getLogger('')

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(console_handler)

    if args.log_level:
        logger.setLevel(args.log_level)
        console_handler.setLevel(args.log_level)

    base_path = os.path.dirname(os.path.realpath(__file__))
    icon_path_mask = os.path.join(base_path, "icons", "charts_{}.png")

    app = PyDMApplication(hide_nav_bar=True, hide_menu_bar=True,
                          hide_status_bar=True, use_main_window=False)

    app_icon = QtGui.QIcon()
    app_icon.addFile(icon_path_mask.format(16), QtCore.QSize(16, 16))
    app_icon.addFile(icon_path_mask.format(24), QtCore.QSize(24, 24))
    app_icon.addFile(icon_path_mask.format(32), QtCore.QSize(32, 32))
    app_icon.addFile(icon_path_mask.format(64), QtCore.QSize(64, 64))
    app_icon.addFile(icon_path_mask.format(128), QtCore.QSize(128, 128))
    app_icon.addFile(icon_path_mask.format(256), QtCore.QSize(256, 256))

    app.setWindowIcon(app_icon)

    config_file = None
    if args.config_file:
        config_file = os.path.expandvars(os.path.expanduser(args.config_file))

    display = TimeChartDisplay(config_file=config_file, args=extra_args)
    display.show()

    sys.exit(app.exec_())


def _parse_arguments():
    """
    Parse the command arguments.

    Returns
    -------
    The command arguments as a dictionary : dict
    """

    parser = argparse.ArgumentParser(
        description="A charting tool based on the Python Display Manager (PyDM).")

    parser.add_argument(
        '--log_level',
        help='Configure level of log display',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        default='INFO'
    )

    parser.add_argument("--config-file",
                        help="The configuration file to import to and start TimeChart.")
    parser.add_argument('--version', action='version',
                        version='TimeChart {version}'.format(
                            version=timechart.__version__))

    args, extra_args = parser.parse_known_args()
    return args, extra_args


if __name__ == "__main__":
    try:
        main()
    except Exception as error:
        traceback.print_exc()



"""
Configuration Data Importing, for Both TimeChart and StripTool Configuration Files
"""

import six
import json

from .importers.timechart_config_importer import TimeChartConfigImporter
from .importers.striptool_config_importer import StripToolConfigImporter
from ..utilities.utils import serialize_colors


import logging
logger = logging.getLogger(__name__)


class SettingsImporterException(Exception):
    pass


class SettingsImporter:
    """
    Import configuration data from files exported from different sources, i.e. TimeChart and
    StripTool.
    """
    def __init__(self, pydm_main_display=None):
        """
        Parameters
        ----------
        pydm_main_display : PyDMDisplay
            The Main Window object.
        """
        self.timechart_importer = TimeChartConfigImporter(pydm_main_display)
        self.striptool_importer = StripToolConfigImporter(pydm_main_display)

    def import_settings(self, filename):
        """
        Import the settings from a file, which could be a TimeChart JSON file, or a StripTool .stp file.

        Parameters
        ----------
        filename : The path to a configuration data file.
        """
        try:
            with open(filename, 'r') as settings_file:
                if filename.endswith(".stp"):
                    logger.warning("The StripTool config file format will soon be unsupported. You can convert this "
                                   "file to the TimeChart config format by clicking on the Export button, then select "
                                   "Chart Settings while running TimeChart.")
                    timechart_settings = self.convert_stp_file(settings_file)
                else:
                    timechart_settings = json.load(settings_file)

            # Apply the settings as TimeChart settings, now that we have all the file data converted to the same
            # TimeChart config format
            self.timechart_importer.apply_settings(timechart_settings)
        except Exception as error:
            six.raise_from(SettingsImporterException(str(error)), error)

    def convert_stp_file(self, stp_file_handle, new_timechart_file=None):
        """
        Convert a StripTool STP file into the TimeChart config data, and write to a TimeChart JSON config file if
        requested.

        Parameters
        ----------
        stp_file_handle : io
            The handle to the STP config file
        new_timechart_file : str
            The full path to the TimeChart file to write the TimeChart config dict to.

        Returns
        -------
            The TimeChart settings dictionary.

        """
        # Parse the StripTool file
        stp_data = self.striptool_importer.import_to_dict(stp_file_handle)

        # Convert to the equivalent TimeChart settings
        timechart_settings = self.striptool_importer.convert_to_timechart_setting(stp_data)

        if new_timechart_file:
            serialize_colors(timechart_settings)

            with open(new_timechart_file, 'w') as output_file:
                json.dump(timechart_settings, fp=output_file, indent=4, separators=(',', ': '))

        return timechart_settings



"""
The Main Display Window
"""

import os
import logging

from functools import partial
import datetime

from pyqtgraph import TextItem

from qtpy.QtCore import Qt, Slot, QTimer
from qtpy.QtWidgets import (QApplication, QWidget, QCheckBox, QHBoxLayout,
                            QVBoxLayout, QFormLayout, QLabel, QSplitter,
                            QComboBox, QLineEdit, QPushButton, QSlider,
                            QSpinBox, QTabWidget,
                            QColorDialog, QGroupBox, QRadioButton,
                            QMessageBox, QFileDialog, QScrollArea, QFrame,
                            QSizePolicy, QLayout,
                            QToolButton, QFontDialog)
from qtpy.QtGui import QColor, QPalette

from pydm import Display
from pydm.widgets.timeplot import (PyDMTimePlot, DEFAULT_X_MIN,
                                   MINIMUM_BUFFER_SIZE, DEFAULT_BUFFER_SIZE)
from pydm.utilities.iconfont import IconFont
from ..data_io.settings_importer import SettingsImporter, SettingsImporterException


from .curve_settings_display import CurveSettingsDisplay
from .axis_settings_display import AxisSettingsDisplay
from .chart_data_export_display import ChartDataExportDisplay
from ..utilities.utils import random_color, display_message_box

from .defaults import *
logger = logging.getLogger(__name__)


class TimeChartDisplay(Display):
    def __init__(self, parent=None, args=[], macros=None, show_pv_add_panel=True, config_file=None):
        """
        Create all the widgets, including any child dialogs.

        Parameters
        ----------
        parent : QWidget
            The parent widget of the charting display
        args : list
            The command parameters
        macros : str
            Macros to modify the UI parameters at runtime
        show_pv_add_panel : bool
            Whether or not to show the PV add panel on top of the graph
        """
        super(TimeChartDisplay, self).__init__(parent=parent, args=args,
                                               macros=macros)
        self.legend_font = None
        self.channel_map = dict()
        self.setWindowTitle("TimeChart Tool")

        self.main_layout = QVBoxLayout()
        self.body_layout = QVBoxLayout()

        self.pv_add_panel = QFrame()
        self.pv_add_panel.setVisible(show_pv_add_panel)
        self.pv_add_panel.setMaximumHeight(50)
        self.pv_layout = QHBoxLayout()
        self.pv_name_line_edt = QLineEdit()
        self.pv_name_line_edt.setAcceptDrops(True)
        self.pv_name_line_edt.returnPressed.connect(self.add_curve)

        self.pv_protocol_cmb = QComboBox()
        self.pv_protocol_cmb.addItems(["ca://", "archive://"])
        self.pv_protocol_cmb.setEnabled(False)

        self.pv_connect_push_btn = QPushButton("Connect")
        self.pv_connect_push_btn.clicked.connect(self.add_curve)

        self.tab_panel = QTabWidget()
        self.tab_panel.setMinimumWidth(350)
        self.tab_panel.setMaximumWidth(350)

        self.curve_settings_tab = QWidget()
        self.data_settings_tab = QWidget()
        self.chart_settings_tab = QWidget()

        self.charting_layout = QHBoxLayout()
        self.chart = PyDMTimePlot(plot_by_timestamps=False)
        self.chart.setDownsampling(ds=False, auto=False, mode=None)
        self.chart.plot_redrawn_signal.connect(self.update_curve_data)
        self.chart.setBufferSize(DEFAULT_BUFFER_SIZE)
        self.chart.setPlotTitle(DEFAULT_CHART_TITLE)

        self.splitter = QSplitter()

        self.curve_settings_layout = QVBoxLayout()
        self.curve_settings_layout.setAlignment(Qt.AlignTop)
        self.curve_settings_layout.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.curve_settings_layout.setSpacing(5)

        self.crosshair_settings_layout = QVBoxLayout()
        self.crosshair_settings_layout.setAlignment(Qt.AlignTop)
        self.crosshair_settings_layout.setSpacing(5)

        self.enable_crosshair_chk = QCheckBox("Crosshair")
        self.crosshair_coord_lbl = QLabel()
        self.crosshair_coord_lbl.setWordWrap(True)

        self.curve_settings_inner_frame = QFrame()
        self.curve_settings_inner_frame.setLayout(self.curve_settings_layout)

        self.curve_settings_scroll = QScrollArea()
        self.curve_settings_scroll.setVerticalScrollBarPolicy(
            Qt.ScrollBarAsNeeded)
        self.curve_settings_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.curve_settings_scroll.setWidget(self.curve_settings_inner_frame)
        self.curve_settings_scroll.setWidgetResizable(True)

        self.enable_crosshair_chk.setChecked(False)
        self.enable_crosshair_chk.clicked.connect(
            self.handle_enable_crosshair_checkbox_clicked)
        self.enable_crosshair_chk.clicked.emit(False)

        self.curves_tab_layout = QHBoxLayout()
        self.curves_tab_layout.addWidget(self.curve_settings_scroll)

        self.data_tab_layout = QVBoxLayout()
        self.data_tab_layout.setAlignment(Qt.AlignTop)
        self.data_tab_layout.setSpacing(5)

        self.chart_settings_layout = QVBoxLayout()
        self.chart_settings_layout.setAlignment(Qt.AlignTop)
        self.chart_settings_layout.setSpacing(5)

        self.chart_layout = QVBoxLayout()
        self.chart_layout.setSpacing(10)

        self.chart_panel = QWidget()
        self.chart_panel.setMinimumHeight(400)

        self.chart_control_layout = QHBoxLayout()
        self.chart_control_layout.setAlignment(Qt.AlignHCenter)
        self.chart_control_layout.setSpacing(10)
        self.zoom_x_layout = QVBoxLayout()
        self.zoom_x_layout.setAlignment(Qt.AlignTop)
        self.zoom_x_layout.setSpacing(5)

        self.plus_icon = IconFont().icon("plus", color=QColor("green"))
        self.minus_icon = IconFont().icon("minus", color=QColor("red"))
        self.view_all_icon = IconFont().icon("globe", color=QColor("blue"))
        self.reset_icon = IconFont().icon("circle-o-notch",
                                          color=QColor("green"))

        self.zoom_in_x_btn = QPushButton("X Zoom")
        self.zoom_in_x_btn.setIcon(self.plus_icon)
        self.zoom_in_x_btn.clicked.connect(
            partial(self.handle_zoom_in_btn_clicked, "x", True))
        self.zoom_in_x_btn.setEnabled(False)

        self.zoom_out_x_btn = QPushButton("X Zoom")
        self.zoom_out_x_btn.setIcon(self.minus_icon)
        self.zoom_out_x_btn.clicked.connect(
            partial(self.handle_zoom_in_btn_clicked, "x", False))
        self.zoom_out_x_btn.setEnabled(False)

        self.zoom_y_layout = QVBoxLayout()
        self.zoom_y_layout.setAlignment(Qt.AlignTop)
        self.zoom_y_layout.setSpacing(5)

        self.zoom_in_y_btn = QPushButton("Y Zoom")
        self.zoom_in_y_btn.setIcon(self.plus_icon)
        self.zoom_in_y_btn.clicked.connect(
            partial(self.handle_zoom_in_btn_clicked, "y", True))
        self.zoom_in_y_btn.setEnabled(False)

        self.zoom_out_y_btn = QPushButton("Y Zoom")
        self.zoom_out_y_btn.setIcon(self.minus_icon)
        self.zoom_out_y_btn.clicked.connect(
            partial(self.handle_zoom_in_btn_clicked, "y", False))
        self.zoom_out_y_btn.setEnabled(False)

        self.view_all_btn = QPushButton("View All")
        self.view_all_btn.setIcon(self.view_all_icon)
        self.view_all_btn.clicked.connect(self.handle_view_all_button_clicked)
        self.view_all_btn.setEnabled(False)

        self.view_all_reset_chart_layout = QVBoxLayout()
        self.view_all_reset_chart_layout.setAlignment(Qt.AlignTop)
        self.view_all_reset_chart_layout.setSpacing(5)

        self.pause_chart_layout = QVBoxLayout()
        self.pause_chart_layout.setAlignment(Qt.AlignTop)
        self.pause_chart_layout.setSpacing(5)

        self.reset_chart_btn = QPushButton("Reset")
        self.reset_chart_btn.setIcon(self.reset_icon)
        self.reset_chart_btn.clicked.connect(
            self.handle_reset_chart_btn_clicked)
        self.reset_chart_btn.setEnabled(False)

        self.pause_icon = IconFont().icon("pause", color=QColor("red"))
        self.play_icon = IconFont().icon("play", color=QColor("green"))
        self.pause_chart_btn = QPushButton()
        self.pause_chart_btn.setIcon(self.pause_icon)
        self.pause_chart_btn.clicked.connect(
            self.handle_pause_chart_btn_clicked)

        self.title_settings_layout = QVBoxLayout()
        self.title_settings_layout.setAlignment(Qt.AlignTop)
        self.title_settings_layout.setSpacing(5)

        self.title_settings_grpbx = QGroupBox("Title and Legend")
        self.title_settings_grpbx.setMaximumHeight(120)

        self.import_export_data_layout = QVBoxLayout()
        self.import_export_data_layout.setAlignment(Qt.AlignTop)
        self.import_export_data_layout.setSpacing(5)

        self.import_data_btn = QPushButton("Import...")
        self.import_data_btn.clicked.connect(
            self.handle_import_data_btn_clicked)

        self.export_data_btn = QPushButton("Export...")
        self.export_data_btn.clicked.connect(
            self.handle_export_data_btn_clicked)

        self.chart_title_layout = QHBoxLayout()
        self.chart_title_layout.setSpacing(10)

        self.chart_title_lbl = QLabel(text="Graph Title")
        self.chart_title_line_edt = QLineEdit()
        self.chart_title_line_edt.setText(self.chart.getPlotTitle())
        self.chart_title_line_edt.textChanged.connect(
            self.handle_title_text_changed)

        self.chart_title_font_btn = QPushButton()
        self.chart_title_font_btn.setFixedHeight(24)
        self.chart_title_font_btn.setFixedWidth(24)
        self.chart_title_font_btn.setIcon(IconFont().icon("font"))
        self.chart_title_font_btn.clicked.connect(
            partial(self.handle_chart_font_changed, "title")
        )

        self.chart_change_axis_settings_btn = QPushButton(
            text="Change Axis Settings...")
        self.chart_change_axis_settings_btn.clicked.connect(
            self.handle_change_axis_settings_clicked)

        self.update_datetime_timer = QTimer(self)
        self.update_datetime_timer.timeout.connect(
            self.handle_update_datetime_timer_timeout)

        self.chart_sync_mode_layout = QVBoxLayout()
        self.chart_sync_mode_layout.setSpacing(5)

        self.chart_sync_mode_grpbx = QGroupBox("Data Sampling Mode")
        self.chart_sync_mode_grpbx.setMaximumHeight(100)

        self.chart_sync_mode_sync_radio = QRadioButton("Synchronous")
        self.chart_sync_mode_async_radio = QRadioButton("Asynchronous")
        self.chart_sync_mode_async_radio.setChecked(True)

        self.graph_drawing_settings_layout = QVBoxLayout()
        self.graph_drawing_settings_layout.setAlignment(Qt.AlignVCenter)

        self.chart_interval_layout = QFormLayout()

        self.chart_redraw_rate_lbl = QLabel("Redraw Rate (Hz)")
        self.chart_redraw_rate_spin = QSpinBox()
        self.chart_redraw_rate_spin.setRange(MIN_REDRAW_RATE_HZ,
                                             MAX_REDRAW_RATE_HZ)
        self.chart_redraw_rate_spin.setValue(DEFAULT_REDRAW_RATE_HZ)
        self.chart_redraw_rate_spin.editingFinished.connect(
            self.handle_redraw_rate_changed)

        self.chart_data_sampling_rate_lbl = QLabel("Data Sampling Rate (Hz)")
        self.chart_data_async_sampling_rate_spin = QSpinBox()
        self.chart_data_async_sampling_rate_spin.setRange(
            MIN_DATA_SAMPLING_RATE_HZ, MAX_DATA_SAMPLING_RATE_HZ)
        self.chart_data_async_sampling_rate_spin.setValue(
            DEFAULT_DATA_SAMPLING_RATE_HZ)
        self.chart_data_async_sampling_rate_spin.editingFinished.connect(
            self.handle_data_sampling_rate_changed)
        self.chart_data_sampling_rate_lbl.hide()
        self.chart_data_async_sampling_rate_spin.hide()

        self.chart_limit_time_span_layout = QHBoxLayout()
        self.chart_limit_time_span_layout.setSpacing(5)

        self.limit_time_plan_text = "Limit Time Span"
        self.chart_limit_time_span_chk = QCheckBox(self.limit_time_plan_text)
        self.chart_limit_time_span_chk.hide()
        self.chart_limit_time_span_lbl = QLabel("Hr:Min:Sec")
        self.chart_limit_time_span_hours_spin_box = QSpinBox()
        self.chart_limit_time_span_hours_spin_box.setMaximum(999)
        self.chart_limit_time_span_minutes_spin_box = QSpinBox()
        self.chart_limit_time_span_minutes_spin_box.setMaximum(59)
        self.chart_limit_time_span_seconds_spin_box = QSpinBox()
        self.chart_limit_time_span_seconds_spin_box.setMaximum(59)
        self.chart_limit_time_span_activate_btn = QPushButton("Apply")
        self.chart_limit_time_span_activate_btn.setDisabled(True)

        self.chart_ring_buffer_layout = QFormLayout()

        self.chart_ring_buffer_size_lbl = QLabel("Ring Buffer Size")
        self.chart_ring_buffer_size_edt = QLineEdit()
        self.chart_ring_buffer_size_edt.returnPressed.connect(
            self.handle_buffer_size_changed)
        self.chart_ring_buffer_size_edt.setText(str(DEFAULT_BUFFER_SIZE))

        self.show_legend_chk = QCheckBox("Show Legend")
        self.show_legend_chk.clicked.connect(
            self.handle_show_legend_checkbox_clicked)
        self.show_legend_chk.setChecked(self.chart.showLegend)

        self.legend_font_btn = QPushButton()
        self.legend_font_btn.setFixedHeight(24)
        self.legend_font_btn.setFixedWidth(24)
        self.legend_font_btn.setIcon(IconFont().icon("font"))
        self.legend_font_btn.clicked.connect(
            partial(self.handle_chart_font_changed, "legend")
        )

        self.graph_background_color_layout = QFormLayout()
        self.axis_grid_color_layout = QFormLayout()

        self.background_color_lbl = QLabel("Graph Background Color ")
        self.background_color_btn = QPushButton()
        self.background_color_btn.setStyleSheet(
            "background-color: " + self.chart.getBackgroundColor().name())
        self.background_color_btn.setContentsMargins(10, 0, 5, 5)
        self.background_color_btn.setMaximumWidth(20)
        self.background_color_btn.clicked.connect(
            self.handle_background_color_button_clicked)

        self.axis_settings_layout = QVBoxLayout()
        self.axis_settings_layout.setSpacing(10)

        self.show_x_grid_chk = QCheckBox("Show x Grid")
        self.show_x_grid_chk.setChecked(self.chart.showXGrid)
        self.show_x_grid_chk.clicked.connect(
            self.handle_show_x_grid_checkbox_clicked)

        self.show_y_grid_chk = QCheckBox("Show y Grid")
        self.show_y_grid_chk.setChecked(self.chart.showYGrid)
        self.show_y_grid_chk.clicked.connect(
            self.handle_show_y_grid_checkbox_clicked)

        self.axis_color_lbl = QLabel("Axis and Grid Color")
        self.axis_color_btn = QPushButton()
        self.axis_color_btn.setStyleSheet(
            "background-color: " + DEFAULT_CHART_AXIS_COLOR.name())
        self.axis_color_btn.setContentsMargins(10, 0, 5, 5)
        self.axis_color_btn.setMaximumWidth(20)
        self.axis_color_btn.clicked.connect(
            self.handle_axis_color_button_clicked)

        self.grid_opacity_lbl = QLabel("Grid Opacity")
        self.grid_opacity_lbl.setEnabled(False)

        self.grid_opacity_slr = QSlider(Qt.Horizontal)
        self.grid_opacity_slr.setFocusPolicy(Qt.StrongFocus)
        self.grid_opacity_slr.setRange(0, 10)
        self.grid_opacity_slr.setValue(5)
        self.grid_opacity_slr.setTickInterval(1)
        self.grid_opacity_slr.setSingleStep(1)
        self.grid_opacity_slr.setTickPosition(QSlider.TicksBelow)
        self.grid_opacity_slr.valueChanged.connect(
            self.handle_grid_opacity_slider_mouse_release)
        self.grid_opacity_slr.setEnabled(False)

        self.reset_data_settings_btn = QPushButton("Reset Data Settings")
        self.reset_data_settings_btn.clicked.connect(
            self.handle_reset_data_settings_btn_clicked)

        self.reset_chart_settings_btn = QPushButton("Reset Chart Settings")
        self.reset_chart_settings_btn.clicked.connect(
            self.handle_reset_chart_settings_btn_clicked)

        self.curve_checkbox_panel = QWidget()

        self.graph_drawing_settings_grpbx = QGroupBox("Graph Intervals")
        self.graph_drawing_settings_grpbx.setAlignment(Qt.AlignTop)

        self.axis_settings_grpbx = QGroupBox("Graph Appearance")

        self.app = QApplication.instance()
        self.setup_ui()

        self.curve_settings_disp = None
        self.axis_settings_disp = None
        self.chart_data_export_disp = None
        self.chart_data_import_disp = None
        self.grid_alpha = 5
        self.time_span_limit_hours = None
        self.time_span_limit_minutes = None
        self.time_span_limit_seconds = None
        self.data_sampling_mode = ASYNC_DATA_SAMPLING

        # If there is an imported config file, let's start TimeChart with the imported configuration data
        if config_file:
            importer = SettingsImporter(self)
            try:
                importer.import_settings(config_file)
            except SettingsImporterException:
                display_message_box(QMessageBox.Critical, "Import Failure",
                                    "Cannot import the file '{0}'. Check the log for the error details."
                                    .format(config_file))
                logger.exception("Cannot import the file '{0}'.".format(config_file))

    def ui_filepath(self):
        """
        The path to the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def ui_filename(self):
        """
        The name the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def setup_ui(self):
        """
        Initialize the widgets and layouts.
        """
        self.setLayout(self.main_layout)

        self.pv_layout.addWidget(self.pv_protocol_cmb)
        self.pv_layout.addWidget(self.pv_name_line_edt)
        self.pv_layout.addWidget(self.pv_connect_push_btn)
        self.pv_add_panel.setLayout(self.pv_layout)
        QTimer.singleShot(0, self.pv_name_line_edt.setFocus)

        self.curve_settings_tab.setLayout(self.curves_tab_layout)

        self.chart_settings_tab.setLayout(self.chart_settings_layout)
        self.setup_chart_settings_layout()

        self.data_settings_tab.setLayout(self.data_tab_layout)
        self.setup_data_tab_layout()

        self.tab_panel.addTab(self.curve_settings_tab, "Curves")
        self.tab_panel.addTab(self.data_settings_tab, "Data")
        self.tab_panel.addTab(self.chart_settings_tab, "Graph")

        self.crosshair_settings_layout.addWidget(self.enable_crosshair_chk)
        self.crosshair_settings_layout.addWidget(self.crosshair_coord_lbl)

        self.zoom_x_layout.addWidget(self.zoom_in_x_btn)
        self.zoom_x_layout.addWidget(self.zoom_out_x_btn)

        self.zoom_y_layout.addWidget(self.zoom_in_y_btn)
        self.zoom_y_layout.addWidget(self.zoom_out_y_btn)

        self.view_all_reset_chart_layout.addWidget(self.reset_chart_btn)
        self.view_all_reset_chart_layout.addWidget(self.view_all_btn)

        self.pause_chart_layout.addWidget(self.pause_chart_btn)

        self.import_export_data_layout.addWidget(self.import_data_btn)
        self.import_export_data_layout.addWidget(self.export_data_btn)

        self.chart_control_layout.addLayout(self.zoom_x_layout)
        self.chart_control_layout.addLayout(self.zoom_y_layout)
        self.chart_control_layout.addLayout(self.view_all_reset_chart_layout)
        self.chart_control_layout.addLayout(self.pause_chart_layout)
        self.chart_control_layout.addLayout(self.crosshair_settings_layout)
        self.chart_control_layout.addLayout(self.import_export_data_layout)
        self.chart_control_layout.insertSpacing(5, 30)

        self.chart_layout.addWidget(self.chart)
        self.chart_layout.addLayout(self.chart_control_layout)

        self.chart_panel.setLayout(self.chart_layout)

        self.splitter.addWidget(self.chart_panel)
        self.splitter.addWidget(self.tab_panel)
        self.splitter.setSizes([1, 0])

        self.splitter.setHandleWidth(10)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)

        self.charting_layout.addWidget(self.splitter)

        self.body_layout.addWidget(self.pv_add_panel)
        self.body_layout.addLayout(self.charting_layout)
        self.body_layout.setSpacing(0)
        self.body_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addLayout(self.body_layout)

        self.enable_chart_control_buttons(False)

        handle = self.splitter.handle(1)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        button = QToolButton(handle)
        button.setArrowType(Qt.LeftArrow)
        button.clicked.connect(
            lambda: self.handle_splitter_button(True))
        layout.addWidget(button)
        button = QToolButton(handle)
        button.setArrowType(Qt.RightArrow)
        button.clicked.connect(
            lambda: self.handle_splitter_button(False))
        layout.addWidget(button)
        handle.setLayout(layout)

    def handle_splitter_button(self, left=True):
        if left:
            self.splitter.setSizes([1, 1])
        else:
            self.splitter.setSizes([1, 0])

    def change_legend_font(self, font):
        if font is None:
            return
        self.legend_font = font
        items = self.chart.plotItem.legend.items
        for i in items:
            i[1].item.setFont(font)
            i[1].resizeEvent(None)
            i[1].updateGeometry()

    def change_title_font(self, font):
        current_text = self.chart.plotItem.titleLabel.text
        args = {
            "family": font.family,
            "size": "{}pt".format(font.pointSize()),
            "bold": font.bold(),
            "italic": font.italic(),
        }
        self.chart.plotItem.titleLabel.setText(current_text, **args)

    def handle_chart_font_changed(self, target):
        if target not in ("title", "legend"):
            return

        dialog = QFontDialog(self)
        dialog.setOption(QFontDialog.DontUseNativeDialog, True)

        if target == "title":
            dialog.fontSelected.connect(self.change_title_font)
        else:
            dialog.fontSelected.connect(self.change_legend_font)

        dialog.open()

    def setup_data_tab_layout(self):
        self.chart_sync_mode_sync_radio.toggled.connect(
            partial(self.handle_sync_mode_radio_toggle,
                    self.chart_sync_mode_sync_radio))
        self.chart_sync_mode_async_radio.toggled.connect(
            partial(self.handle_sync_mode_radio_toggle,
                    self.chart_sync_mode_async_radio))

        self.chart_sync_mode_layout.addWidget(self.chart_sync_mode_sync_radio)
        self.chart_sync_mode_layout.addWidget(self.chart_sync_mode_async_radio)
        self.chart_sync_mode_grpbx.setLayout(self.chart_sync_mode_layout)

        self.data_tab_layout.addWidget(self.chart_sync_mode_grpbx)

        self.chart_limit_time_span_layout.addWidget(
            self.chart_limit_time_span_lbl)
        self.chart_limit_time_span_layout.addWidget(
            self.chart_limit_time_span_hours_spin_box)
        self.chart_limit_time_span_layout.addWidget(
            self.chart_limit_time_span_minutes_spin_box)
        self.chart_limit_time_span_layout.addWidget(
            self.chart_limit_time_span_seconds_spin_box)
        self.chart_limit_time_span_layout.addWidget(
            self.chart_limit_time_span_activate_btn)

        self.chart_limit_time_span_lbl.hide()
        self.chart_limit_time_span_hours_spin_box.hide()
        self.chart_limit_time_span_minutes_spin_box.hide()
        self.chart_limit_time_span_seconds_spin_box.hide()
        self.chart_limit_time_span_activate_btn.hide()

        self.chart_limit_time_span_hours_spin_box.valueChanged.connect(
            self.handle_time_span_changed)
        self.chart_limit_time_span_minutes_spin_box.valueChanged.connect(
            self.handle_time_span_changed)
        self.chart_limit_time_span_seconds_spin_box.valueChanged.connect(
            self.handle_time_span_changed)

        self.chart_limit_time_span_chk.clicked.connect(
            self.handle_limit_time_span_checkbox_clicked)
        self.chart_limit_time_span_activate_btn.clicked.connect(
            self.handle_chart_limit_time_span_activate_btn_clicked)

        self.chart_interval_layout.addRow(self.chart_redraw_rate_lbl,
                                          self.chart_redraw_rate_spin)
        self.chart_interval_layout.addRow(self.chart_data_sampling_rate_lbl,
                                          self.chart_data_async_sampling_rate_spin)
        self.graph_drawing_settings_layout.addLayout(self.chart_interval_layout)

        self.graph_drawing_settings_layout.addWidget(
            self.chart_limit_time_span_chk)
        self.graph_drawing_settings_layout.addLayout(
            self.chart_limit_time_span_layout)

        self.chart_ring_buffer_layout.addRow(self.chart_ring_buffer_size_lbl,
                                             self.chart_ring_buffer_size_edt)

        self.graph_drawing_settings_layout.addLayout(
            self.chart_ring_buffer_layout)
        self.graph_drawing_settings_grpbx.setLayout(
            self.graph_drawing_settings_layout)

        self.data_tab_layout.addWidget(self.graph_drawing_settings_grpbx)
        self.chart_sync_mode_async_radio.toggled.emit(True)

        self.data_tab_layout.addWidget(self.reset_data_settings_btn)

    def setup_chart_settings_layout(self):
        self.chart_title_layout.addWidget(self.chart_title_lbl)
        self.chart_title_layout.addWidget(self.chart_title_line_edt)
        self.chart_title_layout.addWidget(self.chart_title_font_btn)
        self.title_settings_layout.addLayout(self.chart_title_layout)

        legend_layout = QHBoxLayout()
        legend_layout.addWidget(self.show_legend_chk)
        legend_layout.addWidget(self.legend_font_btn)
        self.title_settings_layout.addLayout(legend_layout)
        self.title_settings_layout.addWidget(
            self.chart_change_axis_settings_btn)
        self.title_settings_grpbx.setLayout(self.title_settings_layout)
        self.chart_settings_layout.addWidget(self.title_settings_grpbx)

        self.graph_background_color_layout.addRow(self.background_color_lbl,
                                                  self.background_color_btn)
        self.axis_settings_layout.addLayout(self.graph_background_color_layout)

        self.axis_grid_color_layout.addRow(self.axis_color_lbl,
                                           self.axis_color_btn)
        self.axis_settings_layout.addLayout(self.axis_grid_color_layout)

        self.axis_settings_layout.addWidget(self.show_x_grid_chk)
        self.axis_settings_layout.addWidget(self.show_y_grid_chk)
        self.axis_settings_layout.addWidget(self.grid_opacity_lbl)
        self.axis_settings_layout.addWidget(self.grid_opacity_slr)

        self.axis_settings_grpbx.setLayout(self.axis_settings_layout)

        self.chart_settings_layout.addWidget(self.axis_settings_grpbx)
        self.chart_settings_layout.addWidget(self.reset_chart_settings_btn)

        self.update_datetime_timer.start(1000)

    def add_curve(self):
        """
        Add a new curve to the chart.
        """
        pv_name = self._get_full_pv_name(self.pv_name_line_edt.text())
        if pv_name and len(pv_name):
            color = random_color(curve_colors_only=True)
            for k, v in self.channel_map.items():
                if color == v.color:
                    color = random_color(curve_colors_only=True)

            self.add_y_channel(pv_name=pv_name, curve_name=pv_name, color=color)
            self.handle_splitter_button(left=True)

    def show_mouse_coordinates(self, x, y):
        self.crosshair_coord_lbl.clear()
        self.crosshair_coord_lbl.setText(
            "x = {0:.3f}\ny = {1:.3f}".format(x, y))

    def handle_enable_crosshair_checkbox_clicked(self, is_checked):
        self.chart.enableCrosshair(is_checked)
        self.crosshair_coord_lbl.setVisible(is_checked)

        self.chart.crosshair_position_updated.connect(
            self.show_mouse_coordinates)

    def add_y_channel(self, pv_name, curve_name, color, line_style=Qt.SolidLine,
                      line_width=2, symbol=None, symbol_size=None, is_visible=True):
        if pv_name in self.channel_map:
            logger.error("'{0}' has already been added.".format(pv_name))
            return

        curve = self.chart.addYChannel(y_channel=pv_name, name=curve_name,
                                       color=color, lineStyle=line_style,
                                       lineWidth=line_width, symbol=symbol,
                                       symbolSize=symbol_size)
        curve.show() if is_visible else curve.hide()

        if self.show_legend_chk.isChecked():
            self.change_legend_font(self.legend_font)
        self.channel_map[pv_name] = curve
        self.generate_pv_controls(pv_name, color)

        self.enable_chart_control_buttons()
        try:
            self.app.add_connection(curve.channel)
        except AttributeError:
            # these methods are not needed on future versions of pydm
            pass

    def generate_pv_controls(self, pv_name, curve_color):
        """
        Generate a set of widgets to manage the appearance of a curve. The set of widgets includes:
            1. A checkbox which shows the curve on the chart if checked, and hide the curve if not
               checked
            2. Three buttons -- Modify..., Focus, and Remove. Modify... will bring up the Curve
               Settings dialog. Focus adjusts the chart's zooming for the current curve.
               Remove will delete the curve from the chart
        Parameters
        ----------
        pv_name: str
            The name of the PV the current curve is being plotted for
        curve_color : QColor
            The color of the curve to paint for the checkbox label to help the user track the curve
            to the checkbox
        """
        individual_curve_layout = QVBoxLayout()

        size_policy = QSizePolicy()
        size_policy.setVerticalPolicy(QSizePolicy.Fixed)
        size_policy.setHorizontalPolicy(QSizePolicy.Fixed)

        individual_curve_grpbx = QGroupBox()
        individual_curve_grpbx.setMinimumWidth(300)
        individual_curve_grpbx.setMinimumHeight(120)
        individual_curve_grpbx.setAlignment(Qt.AlignTop)

        individual_curve_grpbx.setSizePolicy(size_policy)

        individual_curve_grpbx.setObjectName(pv_name + "_grb")
        individual_curve_grpbx.setLayout(individual_curve_layout)

        checkbox = QCheckBox(parent=individual_curve_grpbx)
        checkbox.setObjectName(pv_name + "_chb")

        palette = checkbox.palette()
        palette.setColor(QPalette.Active, QPalette.WindowText, curve_color)
        checkbox.setPalette(palette)

        display_name = pv_name.split("://")[1]
        if len(display_name) > MAX_DISPLAY_PV_NAME_LENGTH:
            # Only display max allowed number of characters of the PV Name
            display_name = display_name[
                           :int(MAX_DISPLAY_PV_NAME_LENGTH / 2) - 1] + "..." + \
                           display_name[
                           -int(MAX_DISPLAY_PV_NAME_LENGTH / 2) + 2:]

        checkbox.setText(display_name)

        data_text = QLabel(parent=individual_curve_grpbx)
        data_text.setWordWrap(True)
        data_text.setObjectName(pv_name + "_lbl")
        data_text.setPalette(palette)

        checkbox.setChecked(True)
        checkbox.toggled.connect(partial(self.handle_curve_chkbox_toggled, checkbox))
        if not self.chart.findCurve(pv_name).isVisible():
            checkbox.setChecked(False)

        modify_curve_btn = QPushButton("Modify...",
                                       parent=individual_curve_grpbx)
        modify_curve_btn.setObjectName(pv_name + "_btn_modify")
        modify_curve_btn.setMaximumWidth(80)
        modify_curve_btn.clicked.connect(
            partial(self.display_curve_settings_dialog, pv_name))

        focus_curve_btn = QPushButton("Focus", parent=individual_curve_grpbx)
        focus_curve_btn.setObjectName(pv_name + "_btn_focus")
        focus_curve_btn.setMaximumWidth(80)
        focus_curve_btn.clicked.connect(partial(self.focus_curve, pv_name))

        clear_curve_btn = QPushButton("Clear", parent=individual_curve_grpbx)
        clear_curve_btn.setObjectName(pv_name + "_btn_clear")
        clear_curve_btn.setMaximumWidth(80)
        clear_curve_btn.clicked.connect(partial(self.clear_curve, pv_name))

        # annotate_curve_btn = QPushButton("Annotate...",
        #                                  parent=individual_curve_grpbx)
        # annotate_curve_btn.setObjectName(pv_name+"_btn_ann")
        # annotate_curve_btn.setMaximumWidth(80)
        # annotate_curve_btn.clicked.connect(
        #     partial(self.annotate_curve, pv_name))

        remove_curve_btn = QPushButton("Remove", parent=individual_curve_grpbx)
        remove_curve_btn.setObjectName(pv_name + "_btn_remove")
        remove_curve_btn.setMaximumWidth(80)
        remove_curve_btn.clicked.connect(partial(self.remove_curve, pv_name))

        curve_btn_layout = QHBoxLayout()
        curve_btn_layout.setSpacing(5)
        curve_btn_layout.addWidget(modify_curve_btn)
        curve_btn_layout.addWidget(focus_curve_btn)
        curve_btn_layout.addWidget(clear_curve_btn)
        # curve_btn_layout.addWidget(annotate_curve_btn)
        curve_btn_layout.addWidget(remove_curve_btn)

        individual_curve_layout.addWidget(checkbox)
        individual_curve_layout.addWidget(data_text)
        individual_curve_layout.addLayout(curve_btn_layout)

        self.curve_settings_layout.addWidget(individual_curve_grpbx)

        self.tab_panel.setCurrentIndex(0)

    def handle_curve_chkbox_toggled(self, checkbox):
        """
        Handle a checkbox's checked and unchecked events.

        If a checkbox is checked, find the curve from the channel map. If found, re-draw the curve with its previous
        appearance settings.

        If a checkbox is unchecked, remove the curve from the chart, but keep the cached data in the channel map.

        Parameters
        ----------
        checkbox : QCheckBox
            The current checkbox being toggled
        """
        pv_name = self._get_full_pv_name(checkbox.text())

        if checkbox.isChecked():
            curve = self.channel_map.get(pv_name, None)
            if curve:
                curve.show()
                self.chart.addLegendItem(curve, pv_name,
                                         self.show_legend_chk.isChecked())
                self.change_legend_font(self.legend_font)
        else:
            curve = self.chart.findCurve(pv_name)
            if curve:
                curve.hide()
                self.chart.removeLegendItem(pv_name)

    def display_curve_settings_dialog(self, pv_name):
        """
        Bring up the Curve Settings dialog to modify the appearance of a curve.

        Parameters
        ----------
        pv_name : str
            The name of the PV the curve is being plotted for

        """
        self.curve_settings_disp = CurveSettingsDisplay(self, pv_name)
        self.curve_settings_disp.show()

    def focus_curve(self, pv_name):
        curve = self.chart.findCurve(pv_name)
        if curve:
            self.chart.plotItem.setYRange(curve.minY, curve.maxY, padding=0)

    def clear_curve(self, pv_name):
        curve = self.chart.findCurve(pv_name)
        if curve:
            curve.initialize_buffer()

    def annotate_curve(self, pv_name):
        curve = self.chart.findCurve(pv_name)
        if curve:
            annot = TextItem(
                html='<div style="text-align: center"><span style="color: #FFF;">This is the'
                     '</span><br><span style="color: #FF0; font-size: 16pt;">PEAK</span></div>',
                anchor=(-0.3, 0.5), border='w', fill=(0, 0, 255, 100))
            self.chart.annotateCurve(curve, annot)

    def remove_curve(self, pv_name):
        """
        Remove a curve from the chart permanently. This will also clear the channel map cache from retaining the
        removed curve's appearance settings.

        Parameters
        ----------
        pv_name : str
            The name of the PV the curve is being plotted for
        """
        curve = self.chart.findCurve(pv_name)
        if curve:
            try:
                self.app.remove_connection(curve.channel)
            except AttributeError:
                # these methods are not needed on future versions of pydm
                pass
            self.chart.removeYChannel(curve)
            del self.channel_map[pv_name]
            self.chart.removeLegendItem(pv_name)

            widget = self.findChild(QGroupBox, pv_name + "_grb")
            if widget:
                widget.deleteLater()

        if len(self.chart.getCurves()) < 1:
            self.enable_chart_control_buttons(False)
            self.show_legend_chk.setChecked(False)

    def handle_title_text_changed(self, new_text):
        self.chart.setPlotTitle(new_text)

    def handle_change_axis_settings_clicked(self):
        self.axis_settings_disp = AxisSettingsDisplay(self)
        self.axis_settings_disp.show()

    def handle_limit_time_span_checkbox_clicked(self, is_checked):
        self.chart_limit_time_span_lbl.setVisible(is_checked)
        self.chart_limit_time_span_hours_spin_box.setVisible(is_checked)
        self.chart_limit_time_span_minutes_spin_box.setVisible(is_checked)
        self.chart_limit_time_span_seconds_spin_box.setVisible(is_checked)
        self.chart_limit_time_span_activate_btn.setVisible(is_checked)

        self.chart_ring_buffer_size_lbl.setDisabled(is_checked)
        self.chart_ring_buffer_size_edt.setDisabled(is_checked)

        if not is_checked:
            self.chart_limit_time_span_chk.setText(self.limit_time_plan_text)

    def handle_time_span_changed(self):
        self.time_span_limit_hours = self.chart_limit_time_span_hours_spin_box.value()
        self.time_span_limit_minutes = self.chart_limit_time_span_minutes_spin_box.value()
        self.time_span_limit_seconds = self.chart_limit_time_span_seconds_spin_box.value()

        status = self.time_span_limit_hours > 0 or self.time_span_limit_minutes > 0 or self.time_span_limit_seconds > 0

        self.chart_limit_time_span_activate_btn.setEnabled(status)

    def handle_chart_limit_time_span_activate_btn_clicked(self):
        timeout_milliseconds = (self.time_span_limit_hours * 3600 + self.time_span_limit_minutes * 60 +
                                self.time_span_limit_seconds) * 1000
        self.chart.setTimeSpan(timeout_milliseconds / 1000.0)
        self.chart_ring_buffer_size_edt.setText(str(self.chart.getBufferSize()))

    def handle_buffer_size_changed(self):
        try:
            new_buffer_size = int(self.chart_ring_buffer_size_edt.text())
            if new_buffer_size and int(new_buffer_size) >= MINIMUM_BUFFER_SIZE:
                self.chart.setBufferSize(new_buffer_size)
        except ValueError:
            display_message_box(QMessageBox.Critical, "Invalid Values",
                                "Only integer values are accepted.")

    def handle_redraw_rate_changed(self):
        self.chart.maxRedrawRate = self.chart_redraw_rate_spin.value()

    def handle_data_sampling_rate_changed(self):
        # The chart expects the value in milliseconds
        sampling_rate_seconds = 1.0 / self.chart_data_async_sampling_rate_spin.value()
        buffer_size = self.chart.getBufferSize()
        self.chart.setUpdateInterval(sampling_rate_seconds)
        if self.chart.getBufferSize() < buffer_size:
            self.chart.setBufferSize(buffer_size)
        self.chart_ring_buffer_size_edt.setText(str(self.chart.getBufferSize()))

    def handle_background_color_button_clicked(self):
        selected_color = QColorDialog.getColor()
        self.chart.setBackgroundColor(selected_color)
        self.background_color_btn.setStyleSheet(
            "background-color: " + selected_color.name())

    def handle_axis_color_button_clicked(self):
        selected_color = QColorDialog.getColor()
        self.chart.setAxisColor(selected_color)
        self.axis_color_btn.setStyleSheet(
            "background-color: " + selected_color.name())

    def handle_grid_opacity_slider_mouse_release(self):
        self.grid_alpha = float(self.grid_opacity_slr.value()) / 10.0
        self.chart.setShowXGrid(self.show_x_grid_chk.isChecked(),
                                self.grid_alpha)
        self.chart.setShowYGrid(self.show_y_grid_chk.isChecked(),
                                self.grid_alpha)

    def handle_show_x_grid_checkbox_clicked(self, is_checked):
        self.chart.setShowXGrid(is_checked, self.grid_alpha)
        self.grid_opacity_lbl.setEnabled(
            is_checked or self.show_y_grid_chk.isChecked())
        self.grid_opacity_slr.setEnabled(
            is_checked or self.show_y_grid_chk.isChecked())

    def handle_show_y_grid_checkbox_clicked(self, is_checked):
        self.chart.setShowYGrid(is_checked, self.grid_alpha)
        self.grid_opacity_lbl.setEnabled(
            is_checked or self.show_x_grid_chk.isChecked())
        self.grid_opacity_slr.setEnabled(
            is_checked or self.show_x_grid_chk.isChecked())

    def handle_show_legend_checkbox_clicked(self, is_checked):
        self.chart.setShowLegend(is_checked)

    def handle_export_data_btn_clicked(self):
        self.chart_data_export_disp = ChartDataExportDisplay(self)
        self.chart_data_export_disp.show()

    def handle_import_data_btn_clicked(self):
        open_file_info = QFileDialog.getOpenFileName(self, caption="Open File", directory=os.path.expanduser('~'),
                                                     filter=IMPORT_FILE_FORMAT)
        open_filename = open_file_info[0]
        if open_filename:
            try:
                importer = SettingsImporter(self)
                importer.import_settings(open_filename)
            except SettingsImporterException:
                display_message_box(QMessageBox.Critical, "Import Failure",
                                    "Cannot import the file '{0}'. Check the log for the error details."
                                    .format(open_filename))
                logger.exception("Cannot import the file '{0}'".format(open_filename))

    def handle_sync_mode_radio_toggle(self, radio_btn):
        if radio_btn.isChecked():
            if radio_btn.text() == "Synchronous":
                self.data_sampling_mode = SYNC_DATA_SAMPLING

                self.chart_data_sampling_rate_lbl.hide()
                self.chart_data_async_sampling_rate_spin.hide()

                self.chart.resetTimeSpan()
                self.chart_limit_time_span_chk.setChecked(False)
                self.chart_limit_time_span_chk.clicked.emit(False)
                self.chart_limit_time_span_chk.hide()

                self.chart.setUpdatesAsynchronously(False)
            elif radio_btn.text() == "Asynchronous":
                self.data_sampling_mode = ASYNC_DATA_SAMPLING

                self.chart_data_sampling_rate_lbl.show()
                self.chart_data_async_sampling_rate_spin.show()
                self.chart_limit_time_span_chk.show()

                self.chart.setUpdatesAsynchronously(True)

    def handle_zoom_in_btn_clicked(self, axis, is_zoom_in):
        scale_factor = 0.5
        if not is_zoom_in:
            scale_factor += 1.0
        if axis == "x":
            self.chart.getViewBox().scaleBy(x=scale_factor)
        elif axis == "y":
            self.chart.getViewBox().scaleBy(y=scale_factor)

    def handle_view_all_button_clicked(self):
        self.chart.plotItem.getViewBox().autoRange()

    def handle_pause_chart_btn_clicked(self):
        if self.chart.pausePlotting():
            self.pause_chart_btn.setIcon(self.pause_icon)
        else:
            self.pause_chart_btn.setIcon(self.play_icon)

    def handle_reset_chart_btn_clicked(self):
        self.chart.getViewBox().setXRange(DEFAULT_X_MIN, 0)
        self.chart.resetAutoRangeY()

    @Slot()
    def handle_reset_chart_settings_btn_clicked(self):
        self.chart.setBackgroundColor(DEFAULT_CHART_BACKGROUND_COLOR)
        self.background_color_btn.setStyleSheet(
            "background-color: " + DEFAULT_CHART_BACKGROUND_COLOR.name())

        self.chart.setAxisColor(DEFAULT_CHART_AXIS_COLOR)
        self.axis_color_btn.setStyleSheet(
            "background-color: " + DEFAULT_CHART_AXIS_COLOR.name())

        self.grid_opacity_slr.setValue(5)

        self.show_x_grid_chk.setChecked(False)
        self.show_x_grid_chk.clicked.emit(False)

        self.show_y_grid_chk.setChecked(False)
        self.show_y_grid_chk.clicked.emit(False)

        self.show_legend_chk.setChecked(False)

        self.chart.setShowXGrid(False)
        self.chart.setShowYGrid(False)
        self.chart.setShowLegend(False)

    @Slot()
    def handle_reset_data_settings_btn_clicked(self):
        self.chart_ring_buffer_size_edt.setText(str(DEFAULT_BUFFER_SIZE))

        self.chart_redraw_rate_spin.setValue(DEFAULT_REDRAW_RATE_HZ)
        self.handle_redraw_rate_changed()

        self.chart_data_async_sampling_rate_spin.setValue(
            DEFAULT_DATA_SAMPLING_RATE_HZ)
        self.chart_data_sampling_rate_lbl.hide()
        self.chart_data_async_sampling_rate_spin.hide()

        self.chart_sync_mode_async_radio.setChecked(True)
        self.chart_sync_mode_async_radio.toggled.emit(True)

        self.chart_limit_time_span_chk.setChecked(False)
        self.chart_limit_time_span_chk.setText(self.limit_time_plan_text)
        self.chart_limit_time_span_chk.clicked.emit(False)

        self.chart.setUpdatesAsynchronously(True)
        self.chart.resetTimeSpan()
        self.chart.resetUpdateInterval()
        self.chart.setBufferSize(DEFAULT_BUFFER_SIZE)

    def enable_chart_control_buttons(self, enabled=True):
        self.zoom_in_x_btn.setEnabled(enabled)
        self.zoom_out_x_btn.setEnabled(enabled)
        self.zoom_in_y_btn.setEnabled(enabled)
        self.zoom_out_y_btn.setEnabled(enabled)

        self.view_all_btn.setEnabled(enabled)
        self.reset_chart_btn.setEnabled(enabled)
        self.pause_chart_btn.setIcon(self.pause_icon)
        self.pause_chart_btn.setEnabled(enabled)
        self.export_data_btn.setEnabled(enabled)

    def _get_full_pv_name(self, pv_name):
        """
        Append the protocol to the PV Name.

        Parameters
        ----------
        pv_name : str
            The name of the PV the curve is being plotted for
        """
        if pv_name and "://" not in pv_name:
            pv_name = ''.join([self.pv_protocol_cmb.currentText(), pv_name])
        return pv_name

    def handle_update_datetime_timer_timeout(self):
        current_label = self.chart.getBottomAxisLabel()
        new_label = "Current Time: " + TimeChartDisplay.get_current_datetime()

        if X_AXIS_LABEL_SEPARATOR in current_label:
            current_label = current_label[
                            current_label.find(X_AXIS_LABEL_SEPARATOR) + len(
                                X_AXIS_LABEL_SEPARATOR):]
            new_label += X_AXIS_LABEL_SEPARATOR + current_label

        self.chart.setLabel("bottom", text=new_label)

    def update_curve_data(self, curve):
        """
        Determine if the PV is active. If not, disable the related PV controls.
        If the PV is active, update the PV controls' states.

        Parameters
        ----------
        curve : PlotItem
           A PlotItem, i.e. a plot, to draw on the chart.
        """
        pv_name = curve.address
        min_y = curve.minY if curve.minY else 0
        max_y = curve.maxY if curve.maxY else 0
        current_y = curve.data_buffer[1, -1]

        grb = self.findChild(QGroupBox, pv_name + "_grb")

        lbl = grb.findChild(QLabel, pv_name + "_lbl")
        lbl.setText("(yMin = {0:.3f}, yMax = {1:.3f}) y = {2:.3f}".format(
            min_y, max_y, current_y))

        chb = grb.findChild(QCheckBox, pv_name + "_chb")

        connected = curve.connected
        if connected and chb.isEnabled():
            return

        chb.setEnabled(connected)
        btn_modify = grb.findChild(QPushButton, pv_name + "_btn_modify")
        btn_modify.setEnabled(connected)
        btn_focus = grb.findChild(QPushButton, pv_name + "_btn_focus")
        btn_focus.setEnabled(connected)

        # btn_ann = grb.findChild(QPushButton, pv_name + "_btn_ann")
        # btn_ann.setEnabled(connected)

    @staticmethod
    def get_current_datetime():
        current_date = datetime.datetime.now().strftime("%b %d, %Y")
        current_time = datetime.datetime.now().strftime("%H:%M:%S")
        current_datetime = current_time + ' (' + current_date + ')'

        return current_datetime

    @property
    def gridAlpha(self):
        return self.grid_alpha



import random

from qtpy.QtCore import Qt
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QMessageBox

from pydm.utilities.colors import svg_color_from_hex

# If TimeChart is importing a StripTool config file, which contains a list of preferred colors, TimeChart will populate
# the STRIPTOOL_PREDEFINED_COLORS list. When TimeChart picks a random color for a curve, it'll pick from
# STRIPTOOL_PREDEFINED_COLORS instead of the predefined colors of TimeChart's own.
#
# On the other hand, if TimeChart imports its own JSON config file, it'll pick a random color for a curve from its
# own color list.
global STRIPTOOL_PREDEFINED_COLORS
STRIPTOOL_PREDEFINED_COLORS = []


PREDEFINED_COLORS = (QColor(Qt.red), QColor(Qt.green),  QColor(Qt.darkRed),  QColor(Qt.blue), QColor(Qt.darkGreen),
                     QColor(Qt.cyan), QColor(Qt.darkCyan), QColor(Qt.magenta), QColor(Qt.darkMagenta),
                     QColor(255, 0, 127), QColor(0, 85, 255), QColor(255, 85, 0), QColor(0, 170, 127),
                     QColor(0, 170, 255), QColor(85, 170, 255), QColor(255, 170, 0), QColor(255, 170, 255))


def add_striptool_color(color):
    STRIPTOOL_PREDEFINED_COLORS.append(color)


def random_color(curve_colors_only=False):
    """
    Return a QColor object from a predefined list.

    Parameters
    ----------
    curve_colors_only : bool
        If True, exclude black or white as the randomized color since a curve could be blended into the background. If
        False, include black or white as one of the randomized colors.

    Returns
    -------
    A color object randomly selected from the predefined list, and is not a white or black color if this is a color
    purported for a curve.
    """
    color_pick = pick_random_color()
    while curve_colors_only and color_pick in (QColor("black"), QColor("white")):
        color_pick = pick_random_color()

    return color_pick


def pick_random_color():
    """
    Straight-up pick a random color, with affinity given to the StripTool-predefined colors.

    Returns : QColor
    -------
    A color object randomly selected from the predefined list.
    """
    if len(STRIPTOOL_PREDEFINED_COLORS) > 0:
        return random.choice(STRIPTOOL_PREDEFINED_COLORS)
    else:
        return random.choice(PREDEFINED_COLORS)


def serialize_colors(timechart_settings):
    """
    Helper method to serialize QColor objects to string, useful for dumping converted contents into
    a JSON file.

    Parameters
    ----------
    timechart_settings : OrderDict
        The dictionary containing the converted configuration data.
    """
    for k, v in timechart_settings.items():
        if isinstance(timechart_settings[k], dict):
            serialize_colors(timechart_settings[k])
        elif "color" in k and v is not None:
            timechart_settings[k] = str(svg_color_from_hex(v.name(), hex_on_fail=True))


def display_message_box(icon, window_title, text):
    """
    Display a message box to the user, usually to inform the user of an invalid input value.

    Parameters
    ----------
    icon : QIcon
        The icon to display in the message box, i.e. Warning, Error, Info, etc.
    window_title : str
        The title of the message box window
    text : str
        The message to show to the user
    """
    msg_box = QMessageBox()
    msg_box.setIcon(icon)
    msg_box.setWindowTitle(window_title)
    msg_box.setText(text)

    msg_box.exec_()




"""
StripTool Configuration Data File Import Strategy
"""

from collections import OrderedDict
import re

from qtpy.QtGui import QColor
from qtpy.QtCore import Qt

from timechart import __version__ as ver

from .timechart_config_importer import TimeChartConfigImporter
from ...utilities.utils import add_striptool_color
from timechart.displays.defaults import DEFAULT_CHART_TITLE, ASYNC_DATA_SAMPLING


def _convert_to_bool(key, chart_settings, value):
    chart_settings[key] = True if int(value) > 0 else False


def _convert_to_int(key, chart_settings, value):
    chart_settings[key] = int(value)


def _convert_to_float(key, chart_settings, value):
    chart_settings[key] = float(value)


def _convert_to_hz(key, chart_settings, time_per_second):
    chart_settings[key] = 1.0 / float(time_per_second)


def _adjust_color_value(color_value):
    try:
        color_value = color_value.rstrip().split(" ")
        color = QColor(int(color_value[0]) // 257, int(color_value[1]) // 257, int(color_value[2]) // 257)

        return color
    except IndexError as error:
        raise error


def _convert_to_qcolor(key, chart_settings, color_value):
    chart_settings[key] = _adjust_color_value(color_value)


def _convert_to_time_span(chart_settings, time_span_seconds):
    time_span_seconds = int(time_span_seconds)
    if time_span_seconds:
        chart_settings["limit_time_span"] = True
        minutes, seconds = divmod(time_span_seconds, 60)
        hours, minutes = divmod(minutes, 60)

        chart_settings["time_span_limit_hours"] = hours
        chart_settings["time_span_limit_minutes"] = minutes
        chart_settings["time_span_limit_seconds"] = seconds


global LINE_WIDTH
LINE_WIDTH = 1


def _add_to_curves(key, value, chart_pvs, index_to_pv_names):
    """
    Form a list of PVs from parsing the StripTool configuration file.
    Then, determine the specific values for each PV to assign to the TimeChart configuration data structure.

    Parameters
    ----------
    key : str
        A key from the StripTool config file
    value : str
        A value paired with a key from the StripTool config file
    chart_pvs : OrderedDict
        A dictionary of PVs, and for each PV, there's a nested dict to hold the PV's specific attribute values
    index_to_pv_names : OrderedDict
        A map between the StripTool's label for each PV, and the PV name.
    """
    key_components = key.split('.')
    key_specific_item = key_components[-1]
    key_index = int(key_components[-2])

    if "Name" in key:
        value = "ca://" + value
        chart_pvs[value] = OrderedDict()

        global LINE_WIDTH
        chart_pvs[value]["line_width"] = LINE_WIDTH

        chart_pvs[value]["y_channel"] = value
        chart_pvs[value]["line_style"] = Qt.SolidLine
        chart_pvs[value]["symbol"] = None
        chart_pvs[value]["symbol_size"] = 1

        index_to_pv_names[key_index] = value
    else:
        pv_name = index_to_pv_names[key_index]
        converted_content = StripToolConfigImporter.CONVERSION_TABLE[key_specific_item]

        if isinstance(converted_content, tuple):
            if converted_content[0]:
                converted_content[1](converted_content[0], chart_pvs[pv_name], value)
            else:
                converted_content[1](chart_pvs[pv_name], value)
        else:
            chart_pvs[pv_name][converted_content] = value


def _customize_colors(value):
    """
    Add the StripTool-defined colors to TimeChart's random color collections.

    Parameters
    ----------
    value : QColor
        A StripTool-defined color, as specified in the StripTool config file.
    """
    add_striptool_color(_adjust_color_value(value))


def _convert_to_line_width(_, __, line_width_value):
    """
    Since StripTool defines the same line width for all curves while TimeChart lets each curve have its own line width,
    here we must import the global StripTool line width, save it to a global variable, and then, whenever it's time
    to provide the line width for a curve, retrieve the global line width variable to assign to that curve.

    Parameters
    ----------
    line_width_value : str
        The line width as set by the StripTool config file.
    """
    global LINE_WIDTH
    LINE_WIDTH = int(line_width_value)


class StripToolConfigImporter(TimeChartConfigImporter):
    """
    The strategy to import a StripTool configuration data file.
    """
    CONVERSION_TABLE = {
        "Units": "unit",
        "Comment": "comment",
        "Min": ("x_min", _convert_to_float),
        "Max": ("x_max", _convert_to_float),
        "Scale": ("scale", _convert_to_int),
        "PlotStatus": ("is_visible", _convert_to_bool),
        "Precision": ("precision", _convert_to_int),
        "Strip.Time.NumSamples": "buffer_size",
        "Strip.Time.SampleInterval": ("update_interval_hz", _convert_to_hz),
        "Strip.Time.Timespan": (None, _convert_to_time_span),
        "Strip.Time.RefreshInterval": ("redraw_rate", _convert_to_hz),
        "Strip.Color.Background": ("background_color", _convert_to_qcolor),
        "Strip.Color.Grid": ("axis_color", _convert_to_qcolor),
        "Strip.Option.GridXon": ("show_x_grid", _convert_to_bool),
        "Strip.Option.GridYon": ("show_y_grid", _convert_to_bool),
        "Strip.Option.GraphLineWidth": ("line_width", _convert_to_line_width)
    }

    def __init__(self, pydm_main_display):
        """
        Parameters
        ----------
        pydm_main_display : PyDMDisplay
            The Main Display Window.
        """
        super(StripToolConfigImporter, self).__init__(pydm_main_display)

    def import_to_dict(self, settings_file):
        """
        Parse the StripTool config file.

        Parameters
        ----------
        settings_file : TextIOWrapper
            The file object to parse

        Returns
        -------
            The file data as a dictionary
        """
        lines = [line.rstrip('\n') for line in settings_file.readlines()]
        stp_data = self._read_file_into_dict(lines)
        return stp_data

    def convert_to_timechart_setting(self, stp_data):
        """
        Convert StripTool settings into TimeChart settings

        Parameters
        ----------
        stp_data : dict
            The StripTool data to convert.

        Returns
        -------
            The equivalent TimeChart settings.

        """
        timechart_settings = OrderedDict()
        timechart_settings["__version__"] = ver

        timechart_settings["pvs"] = OrderedDict()

        timechart_settings["chart_settings"] = OrderedDict()
        timechart_settings["chart_settings"]["title"] = DEFAULT_CHART_TITLE
        timechart_settings["chart_settings"]["x_axis_label"] = None
        timechart_settings["chart_settings"]["x_axis_unit"] = None
        timechart_settings["chart_settings"]["left_y_axis_label"] = None
        timechart_settings["chart_settings"]["left_y_axis_unit"] = None
        timechart_settings["chart_settings"]["right_y_axis_label"] = None
        timechart_settings["chart_settings"]["right_y_axis_unit"] = None
        timechart_settings["chart_settings"]["data_sampling_mode"] = ASYNC_DATA_SAMPLING
        timechart_settings["chart_settings"]["show_legend"] = False
        timechart_settings["chart_settings"]["grid_alpha"] = 5

        index_to_pv_name = OrderedDict()
        for k, v in stp_data.items():
            converted_content = StripToolConfigImporter.CONVERSION_TABLE.get(k, None)

            if "Strip.Curve" in k:
                _add_to_curves(k, v, timechart_settings["pvs"], index_to_pv_name)
            elif "Strip.Color" in k:
                _customize_colors(v)

            if isinstance(converted_content, tuple):
                if converted_content[0]:
                    converted_content[1](converted_content[0], timechart_settings["chart_settings"], v)
                else:
                    converted_content[1](timechart_settings["chart_settings"], v)
            elif converted_content:
                timechart_settings["chart_settings"][converted_content] = v

        return timechart_settings

    def _read_file_into_dict(self, file_contents):
        """
        Parse a file's contents into a dictionary for subsequent filtering.
        For each line, expect two strings, separated by a space. The first string becomes a dictionary entry's key,
        and the next one is the dictionary entry's value.

        Parameters
        ----------
        file_contents : list
            A list of strings, each of which is a line from a text file

        Returns
        -------
            A dictionary with the keys and values parsed from the file contents.
        """
        settings_modules = OrderedDict()
        for line in file_contents:
            tokens = re.sub(' +', ' ', line.lstrip())
            tokens = tokens.split(' ', 1)
            try:
                settings_modules[tokens[0]] = tokens[1]
            except IndexError:
                settings_modules[tokens[0]] = None
        return settings_modules



"""
The Curve Settings Dialog
"""

from pydm import Display
from pydm.widgets.baseplot import BasePlotCurveItem

from qtpy.QtCore import Qt, QSize
from qtpy.QtWidgets import (QFormLayout, QLabel, QComboBox, QSpinBox,
                            QPushButton, QColorDialog, QCheckBox, QGroupBox)
from qtpy.QtGui import QPalette


class CurveSettingsDisplay(Display):
    def __init__(self, main_display, pv_name, parent=None):
        """
        Create all the widgets for the curve appearance settings.

        Parameters
        ----------
        main_display : TimeChartDisplay
            The main display window
        pv_name: str
            The name of the PV the current curve is being plotted for
        parent : QWidget
            The parent widget, if applicable
        """
        super(CurveSettingsDisplay, self).__init__(parent=parent)
        self.main_layout = QFormLayout()
        self.main_display = main_display

        self.chart = self.main_display.chart
        self.pv_name = pv_name
        self.channel_map = self.main_display.channel_map
        self.app = self.main_display.app

        self.curve_original_color = None
        self.curve_color_lbl = QLabel("Curve Color ")
        self.curve_color_btn = QPushButton()
        self.curve_color_btn.setMaximumWidth(20)
        self.curve_color_btn.clicked.connect(
            self.handle_curve_color_button_clicked)

        self.symbol_lbl = QLabel("Symbol")
        self.symbol_cmb = QComboBox()
        self.symbol_size_lbl = QLabel("Symbol Size")
        self.symbol_size_spin = QSpinBox()

        self.line_style_lbl = QLabel("Line Style")
        self.line_style_cmb = QComboBox()
        self.line_width_lbl = QLabel("Line Width")
        self.line_width_spin = QSpinBox()

        self.set_defaults_btn = QPushButton("Reset")
        self.set_defaults_btn.clicked.connect(self.handle_reset_button_clicked)

        self.close_dialog_btn = QPushButton("Close")
        self.close_dialog_btn.clicked.connect(self.handle_close_button_clicked)

        self.setWindowTitle(self.pv_name.split("://")[1])
        self.setFixedSize(QSize(300, 200))
        self.setWindowModality(Qt.ApplicationModal)

        self.setup_ui()

    def ui_filepath(self):
        """
        The path to the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def ui_filename(self):
        """
        The name of the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def setup_ui(self):
        """
        Initialize the widgets and layouts.
        """
        # Populate the values
        for k, _ in BasePlotCurveItem.symbols.items():
            self.symbol_cmb.addItem(k)
        self.symbol_size_spin.setRange(1, 10)

        for k, _ in BasePlotCurveItem.lines.items():
            self.line_style_cmb.addItem(k)
        self.line_width_spin.setRange(1, 5)

        # Set the widget values to the current settings of the current curve
        self.set_widgets_to_current_curve_settings()

        # Connect to slots to handle signal events
        self.symbol_cmb.currentIndexChanged.connect(
            self.handle_symbol_index_changed)
        self.symbol_size_spin.valueChanged.connect(
            self.handle_symbol_size_changed)

        self.line_style_cmb.currentIndexChanged.connect(
            self.handle_line_style_index_changed)
        self.line_width_spin.valueChanged.connect(
            self.handle_line_width_changed)

        # Add widgets to the form layout
        self.main_layout.setSpacing(10)
        self.main_layout.addRow(self.curve_color_lbl, self.curve_color_btn)
        self.main_layout.addRow(self.symbol_lbl, self.symbol_cmb)
        self.main_layout.addRow(self.symbol_size_lbl, self.symbol_size_spin)
        self.main_layout.addRow(self.line_style_lbl, self.line_style_cmb)
        self.main_layout.addRow(self.line_width_lbl, self.line_width_spin)
        self.main_layout.addRow(self.set_defaults_btn, self.close_dialog_btn)

        # Add the form layout to the main layout of this dialog
        self.setLayout(self.main_layout)

    def set_widgets_to_current_curve_settings(self):
        """
        Make the dialog widgets display the current appearance settings of the current curve.
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            self.curve_color_btn.setStyleSheet(
                "background-color: " + curve.color.name())

            self.set_combo_box(self.symbol_cmb, BasePlotCurveItem.symbols,
                               curve.symbol)
            self.set_combo_box(self.line_style_cmb, BasePlotCurveItem.lines,
                               curve.lineStyle)

            self.symbol_size_spin.setValue(curve.symbolSize)
            self.line_width_spin.setValue(curve.lineWidth)

    def set_combo_box(self, combo_box, reference_dict, curve_setting_value):
        """
        Reverse look up for a dictionary key using a dictionary value, and then set that value to a QComboBox widget.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to set the current value for
        reference_dict : dict
            The reference dict to find the key from a value to set the combo box
        curve_setting_value : str
            The setting value to look for from the reference_dict
        """
        for k, v in reference_dict.items():
            if curve_setting_value == v:
                combo_box.setCurrentText(k)
                break

    def handle_curve_color_button_clicked(self):
        selected_color = QColorDialog.getColor()
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            self.curve_original_color = curve.color
            curve.color = selected_color
            self.curve_color_btn.setStyleSheet(
                "background-color: " + curve.color.name())

            self.chart.refreshCurve(curve)
            self.channel_map[self.pv_name] = curve

    def handle_symbol_index_changed(self, selected_index):
        """
        Handle the change in the curve's symbol from a combo box.

        Parameters
        ----------
        selected_index : int
            The currently selected index from the symbol combo box
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            curve.symbol = BasePlotCurveItem.symbols[
                self.symbol_cmb.itemText(selected_index)]
            self.chart.refreshCurve(curve)
            self.channel_map[self.pv_name] = curve

    def handle_symbol_size_changed(self, new_size):
        """
        Handle the symbol size value change from the symbol size spinner.

        Parameters
        ----------
        new_size : int
            The new symbol size set by the user.
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            curve.symbolSize = new_size
            self.chart.refreshCurve(curve)
            self.channel_map[self.pv_name] = curve

    def handle_line_width_changed(self, new_width):
        """
        Handle the symbol size value change from the line width spinner.

        Parameters
        ----------
        new_width: int
            The new line width set by the user.
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            curve.lineWidth = new_width
            self.chart.refreshCurve(curve)
            self.channel_map[self.pv_name] = curve

    def handle_line_style_index_changed(self, selected_index):
        """
        Handle the change in the curve's line style from a combo box.

        Parameters
        ----------
        selected_index : int
            The currently selected index from the line style combo box
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            curve.lineStyle = BasePlotCurveItem.lines[
                self.line_style_cmb.itemText(selected_index)]
            self.chart.refreshCurve(curve)
            self.channel_map[self.pv_name] = curve

    def handle_reset_button_clicked(self):
        """
        Handle the click of the Reset button. This will set all the dialog widgets to the default curve appearance
        settings.
        """
        curve = self.chart.findCurve(self.pv_name)
        if curve:
            if self.curve_original_color:
                curve.color = self.curve_original_color
                self.curve_color_btn.setStyleSheet(
                    "background-color: " + curve.color.name())

            self.symbol_cmb.setCurrentIndex(0)
            self.symbol_size_spin.setValue(10)

            self.line_style_cmb.setCurrentIndex(1)
            self.line_width_spin.setValue(1)

    def closeEvent(self, event):
        self.handle_close_button_clicked()

    def handle_close_button_clicked(self):
        """
        Close the dialog when the Close button is clicked.
        """
        self.close()

        curve = self.chart.findCurve(self.pv_name)
        if curve:
            # Update the widget checkbox text to the current curve color
            widget = self.main_display.findChild(QGroupBox, self.pv_name + "_grb")
            chb = widget.findChild(QCheckBox, self.pv_name + "_chb")
            lbl = widget.findChild(QLabel, self.pv_name + "_lbl")
            widget = [chb, lbl]
            for w in widget:
                palette = w.palette()
                palette.setColor(QPalette.Active, QPalette.WindowText,
                                 curve.color)
                w.setPalette(palette)



"""
Global Constant Definitions
"""

from qtpy.QtGui import QColor


MIN_REDRAW_RATE_HZ = 1
MAX_REDRAW_RATE_HZ = 240
DEFAULT_REDRAW_RATE_HZ = 30

MIN_DATA_SAMPLING_RATE_HZ = 1
MAX_DATA_SAMPLING_RATE_HZ = 360
DEFAULT_DATA_SAMPLING_RATE_HZ = 10

DEFAULT_EXPORTED_IMAGE_WIDTH = "800"
DEFAULT_EXPORTED_IMAGE_HEIGHT = "600"

DEFAULT_CHART_BACKGROUND_COLOR = QColor("black")
DEFAULT_CHART_AXIS_COLOR = QColor("white")
DEFAULT_CHART_TITLE = "TimeChart"

MAX_DISPLAY_PV_NAME_LENGTH = 30

ASYNC_DATA_SAMPLING = 0
SYNC_DATA_SAMPLING = 1

X_AXIS_LABEL_SEPARATOR = " -- "
IMPORT_FILE_FORMAT = "TimeChart or EPICS StripTool file (*.json *.stp)"




"""
The settings importer to process TimeChart configuration data
"""

from qtpy.QtGui import QColor

from ...displays.defaults import ASYNC_DATA_SAMPLING, SYNC_DATA_SAMPLING
from ...utilities.utils import random_color


class TimeChartConfigImporter(object):
    def __init__(self, pydm_main_dislay):
        self.main_display = pydm_main_dislay
        self.chart = self.main_display.chart if self.main_display else None

    def _set_chart_labels(self, chart_labels):
        """
        Set the text and unit for individual labels from a chart label collection.

        Parameters
        ----------
        chart_labels : dict
            A dictionary holding the label position as the key, and a tuple of (text, unit) as the value.
            Either text or unit, or both, will be set as long as each value is not None.
        """
        for label_position, label_data in chart_labels.items():
            text = label_data[0]
            unit = label_data[1]
            if text:
                self.chart.setLabel(label_position, text=text)
            if unit:
                self.chart.setLabel(label_position, unit=unit)

    def _set_chart_values(self, chart_values):
        """
        Set the value for each individual UI widget.
        Parameters
        ----------
        chart_values : dict
            A dictionary containing widgets as the keys, and the widget values as the values.
        """
        for widget, widget_value in chart_values.items():
            widget.setValue(widget_value)

        self.main_display.handle_data_sampling_rate_changed()
        self.main_display.handle_redraw_rate_changed()
        if self.main_display.chart_limit_time_span_chk.isChecked():
            self.main_display.chart_limit_time_span_chk.clicked.emit(True)
            self.main_display.handle_time_span_changed()
            self.main_display.chart_limit_time_span_activate_btn.clicked.emit()

    def _set_chart_checkboxes(self, checked_data):
        """
        Set the "checked" status of individual checkboxes.

        Parameters
        ----------
        checked_data : dict
            A dictionary with checkboxes as the keys, and their corresponding checked/unchecked statuses as the values.
        """
        for widget, widget_value in checked_data.items():
            widget.setChecked(widget_value)

    def _set_chart_colors(self, color_data):
        """
        Paint a color to a chart's component (axis or background) while also painting the background of the button
        widget setting the corresponding component.

        Parameters
        ----------
        color_data : dict
            A dictionary with widgets as the keys, and the color data as the values.
            The color data contains the color to paint, and the functor to apply the color change to a chart's
            component.
        """
        for widget, data in color_data.items():
            color = QColor(data[0])
            set_color_fn = data[1]

            set_color_fn(color)
            widget.setStyleSheet("background-color: " + color.name())

    def apply_settings(self, settings):
        """
        Apply the imported settings to the TimeChart's Main Display.

        Parameters
        ----------
        settings : OrderedDict
            A dictionary of all the imported settings

        self.main_display : PyDMDisplay
            The TimeChart's Main Display
        """        
        pv_settings = settings["pvs"]
        for k, v in pv_settings.items():
            color_value = v.get("color", None)
            color = QColor(color_value) if color_value else random_color(curve_colors_only=True)
            self.main_display.add_y_channel(pv_name=v["y_channel"],
                                            curve_name=k,
                                            color=color,
                                            line_style=v["line_style"],
                                            line_width=v["line_width"],
                                            symbol=v["symbol"],
                                            symbol_size=v["symbol_size"],
                                            is_visible=v.get("is_visible", True))

        chart_settings = settings["chart_settings"]
        if len(chart_settings):
            self.main_display.chart_title_line_edt.textChanged.emit(
                chart_settings["title"])

            chart_labels = {
                "bottom": (chart_settings["x_axis_label"], chart_settings["x_axis_unit"]),
            }
            if chart_settings.get("show_right_y_axis", False):
                chart_labels["right"] = (chart_settings["right_axis_label"], chart_settings["right_axis_unit"]),
            if chart_settings.get("left_axis_label", None):
                chart_labels["left"] = (chart_settings["left_axis_label"], chart_settings["left_axis_unit"]),
            self._set_chart_labels(chart_labels)

            data_sampling_mode = chart_settings["data_sampling_mode"]
            chart_checked_data = {
                self.main_display.chart_sync_mode_sync_radio: data_sampling_mode == SYNC_DATA_SAMPLING,
                self.main_display.chart_sync_mode_async_radio: data_sampling_mode == ASYNC_DATA_SAMPLING,
                self.main_display.chart_limit_time_span_chk: chart_settings["limit_time_span"],
                self.main_display.show_legend_chk: chart_settings["show_legend"],
                self.main_display.show_x_grid_chk: chart_settings["show_x_grid"],
                self.main_display.show_y_grid_chk: chart_settings["show_y_grid"]
            }
            self._set_chart_checkboxes(chart_checked_data)
            self.main_display.show_legend_chk.clicked.emit(chart_settings["show_legend"])
            self.main_display.show_x_grid_chk.clicked.emit(chart_settings["show_x_grid"])
            self.main_display.show_y_grid_chk.clicked.emit(chart_settings["show_y_grid"])

            chart_values = {
                self.main_display.chart_redraw_rate_spin: chart_settings["redraw_rate"],
                self.main_display.chart_data_async_sampling_rate_spin: chart_settings["update_interval_hz"],
                self.main_display.grid_opacity_slr: chart_settings["grid_alpha"],
            }

            time_span_limit_hours = int(chart_settings["time_span_limit_hours"])
            time_span_limit_minutes = int(chart_settings["time_span_limit_minutes"])
            time_span_limit_seconds = int(chart_settings["time_span_limit_seconds"])
            timespan_values = {
                self.main_display.chart_limit_time_span_hours_spin_box: time_span_limit_hours,
                self.main_display.chart_limit_time_span_minutes_spin_box: time_span_limit_minutes,
                self.main_display.chart_limit_time_span_seconds_spin_box: time_span_limit_seconds,
            }
            chart_values.update(timespan_values)
            self._set_chart_values(chart_values)

            self.main_display.chart_ring_buffer_size_edt.setText(str(chart_settings["buffer_size"]))
            self.main_display.handle_buffer_size_changed()

            color_data = {
                self.main_display.background_color_btn: (chart_settings["background_color"],
                                                         self.main_display.chart.setBackgroundColor),
                self.main_display.axis_color_btn: (chart_settings["axis_color"],
                                                   self.main_display.chart.setAxisColor)
            }
            self._set_chart_colors(color_data)



"""
The Dialog to Export Data from a Chart
"""

import os

from qtpy.QtCore import Qt, QSize
from qtpy.QtWidgets import (QVBoxLayout, QFormLayout, QCheckBox, QLineEdit,
                            QFileDialog, QLabel, QComboBox,
                            QPushButton, QColorDialog, QMessageBox)
from qtpy.QtGui import QColor

from pyqtgraph.exporters import CSVExporter, ImageExporter
from pyqtgraph.parametertree import Parameter

from pydm import Display
from ..data_io.settings_exporter import SettingsExporter
from ..utilities.utils import display_message_box

from .defaults import (DEFAULT_EXPORTED_IMAGE_WIDTH, DEFAULT_EXPORTED_IMAGE_HEIGHT)


class TimeChartImageExporter(ImageExporter):
    """
    Override the buggy widthChanged and heightChanged settings from pyqtgraph
    """

    def __init__(self, item):
        super(TimeChartImageExporter, self).__init__(item=item)

    def widthChanged(self):
        sr = self.getSourceRect()
        ar = float(sr.height()) / sr.width()
        self.params.param('height').setValue(int(self.params['width'] * ar))

    def heightChanged(self):
        sr = self.getSourceRect()
        ar = float(sr.width()) / sr.height()
        self.params.param('width').setValue(int(self.params['height'] * ar))


class ChartDataExportDisplay(Display):
    def __init__(self, main_display, parent=None):
        super(ChartDataExportDisplay, self).__init__(parent=parent)
        self.main_layout = QVBoxLayout()
        self.main_layout.setAlignment(Qt.AlignTop)
        self.main_layout.setSpacing(5)
        self.main_display = main_display

        self.export_options_lbl = QLabel()
        self.export_options_lbl.setText("Export Options")
        self.export_options_cmb = QComboBox()
        self.export_options_cmb.addItems(
            ("Curve Data", "Chart Settings", "Image File"))
        self.export_options_cmb.currentIndexChanged.connect(
            self.handle_export_options_index_changed)

        # Options for Chart Settings
        self.include_pv_chk = QCheckBox("Include currently plotted PVs")
        self.include_pv_chk.setChecked(True)

        self.include_chart_settings_chk = QCheckBox(
            "Include current chart settings")
        self.include_chart_settings_chk.setChecked(True)

        # Options for Image File
        self.image_dimension_layout = QFormLayout()
        self.image_dimension_layout.setSpacing(10)

        self.image_width_lbl = QLabel("Image width")
        self.image_width_edt = QLineEdit()
        self.image_width_edt.editingFinished.connect(
            self.handle_image_dimension_value)
        self.image_width_edt.setText(DEFAULT_EXPORTED_IMAGE_WIDTH)

        self.image_height_lbl = QLabel("Image height")
        self.image_height_edt = QLineEdit()
        self.image_height_edt.editingFinished.connect(
            self.handle_image_dimension_value)
        self.image_height_edt.setText(DEFAULT_EXPORTED_IMAGE_HEIGHT)

        self.anti_alias_chk = QCheckBox("Anti-alias")
        self.anti_alias_chk.setChecked(True)

        self.export_image_background_color_lbl = QLabel("Background Color ")
        self.export_image_background_btn = QPushButton()
        self.export_image_background_btn.setMaximumWidth(20)
        self.export_image_background_btn.setStyleSheet(
            "background-color: black")
        self.export_image_background_btn.clicked.connect(
            self.handle_export_image_background_button_clicked)

        self.export_format_lbl = QLabel()
        self.export_format_lbl.setText("Export Format")
        self.file_format_cmb = QComboBox()
        self.file_format_cmb.addItems(("csv", "json", "png"))
        self.file_format = ""
        self.exported_image_background_color = QColor(Qt.black)

        self.save_file_btn = QPushButton("Export...")
        self.save_file_btn.clicked.connect(self.handle_save_file_btn_clicked)

        self.image_width = 0
        self.image_height = 0

        self.setFixedSize(QSize(300, 150))
        self.setWindowTitle("Export Chart Data")
        self.setWindowModality(Qt.ApplicationModal)
        self.setup_ui()

    def ui_filepath(self):
        """
        The path to the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def ui_filename(self):
        """
        The name of the UI file created by Qt Designer, if applicable.
        """
        # No UI file is being used
        return None

    def setup_ui(self):
        self.main_layout.addWidget(self.export_options_lbl)
        self.main_layout.addWidget(self.export_options_cmb)
        self.main_layout.addWidget(self.include_pv_chk)
        self.main_layout.addWidget(self.include_chart_settings_chk)

        self.image_dimension_layout.addRow(self.image_width_lbl,
                                           self.image_width_edt)
        self.image_dimension_layout.addRow(self.image_height_lbl,
                                           self.image_height_edt)
        self.main_layout.addLayout(self.image_dimension_layout)

        self.main_layout.addWidget(self.anti_alias_chk)
        self.main_layout.addWidget(self.export_image_background_color_lbl)
        self.main_layout.addWidget(self.export_image_background_btn)
        self.main_layout.addWidget(self.export_format_lbl)
        self.main_layout.addWidget(self.file_format_cmb)
        self.main_layout.addWidget(self.save_file_btn)
        self.setLayout(self.main_layout)

        self.export_options_cmb.currentIndexChanged.emit(0)

        self.image_width_edt.editingFinished.emit()
        self.image_height_edt.editingFinished.emit()

    def handle_export_options_index_changed(self, selected_index):
        if selected_index == 0:
            self.setFixedSize(QSize(300, 150))
            self.main_layout.setAlignment(Qt.AlignTop)
        elif selected_index == 1:
            self.setFixedSize(QSize(300, 200))
            self.main_layout.setAlignment(Qt.AlignTop)
        elif selected_index == 2:
            self.setFixedSize(QSize(300, 300))
            self.main_layout.setAlignment(Qt.AlignVCenter)

        self.include_pv_chk.setVisible(selected_index == 1)
        self.include_chart_settings_chk.setVisible(selected_index == 1)

        self.image_width_lbl.setVisible(selected_index == 2)
        self.image_width_edt.setVisible(selected_index == 2)
        self.image_height_lbl.setVisible(selected_index == 2)
        self.image_height_edt.setVisible(selected_index == 2)
        self.anti_alias_chk.setVisible(selected_index == 2)
        self.export_image_background_color_lbl.setVisible(selected_index == 2)
        self.export_image_background_btn.setVisible(selected_index == 2)

        self.file_format = ""
        self.file_format_cmb.setCurrentIndex(selected_index)
        supported_formats = self.file_format_cmb.currentText().split(",")
        for f in supported_formats:
            self.file_format += "*." + f + ";;"
        self.file_format = self.file_format[:-2]
        self.file_format_cmb.setEnabled(False)

    def handle_save_file_btn_clicked(self):
        saved_file_info = QFileDialog.getSaveFileName(
            self, caption="Save File", directory=os.path.expanduser('~'),
            filter=self.file_format)
        saved_file_name = saved_file_info[0]
        if saved_file_info[1][1:] not in saved_file_name:
            saved_file_name += saved_file_info[1][1:]

        if saved_file_name:
            if self.export_options_cmb.currentIndex() == 0:
                data_exporter = CSVExporter(self.main_display.chart.plotItem)
                data_exporter.export(saved_file_name)
            elif self.export_options_cmb.currentIndex() == 2:
                image_exporter = TimeChartImageExporter(
                    self.main_display.chart.plotItem)
                image_exporter.params = Parameter(name='params', type='group',
                                                  children=[
                                                      {'name': 'width',
                                                       'type': 'int',
                                                       'value': self.image_width,
                                                       'limits': (0, None)},
                                                      {'name': 'height',
                                                       'type': 'int',
                                                       'value': self.image_height,
                                                       'limits': (0, None)},
                                                      {'name': 'antialias',
                                                       'type': 'bool',
                                                       'value': True},
                                                      {'name': 'background',
                                                       'type': 'color',
                                                       'value': self.exported_image_background_color},
                                                  ])

                image_exporter.widthChanged()
                image_exporter.heightChanged()
                image_exporter.export(fileName=saved_file_name)
            else:
                settings_exporter = SettingsExporter(self.main_display,
                                                     self.include_pv_chk.isChecked(),
                                                     self.include_chart_settings_chk.isChecked())
                settings_exporter.export_settings(saved_file_name)

            self.close()
            QMessageBox.information(self, "Data Export", "Data exported successfully!")

    def handle_export_image_background_button_clicked(self):
        self.exported_image_background_color = QColorDialog.getColor()
        self.export_image_background_btn.setStyleSheet("background-color: " +
                                                       self.exported_image_background_color.name())

    def handle_image_dimension_value(self):
        image_width = self.image_width_edt.text()
        image_height = self.image_height_edt.text()
        try:
            if image_width:
                self.image_width = int(image_width)
                if self.image_width <= 0:
                    raise ValueError
            if image_height:
                self.image_height = int(image_height)
                if self.image_height <= 0:
                    raise ValueError
            elif not image_width and not image_height:
                raise ValueError
        except ValueError:
            display_message_box(QMessageBox.Critical, "Invalid Values",
                                "Only integer values larger than 0 are accepted.")



import versioneer
from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().split()

git_requirements = [r for r in requirements if r.startswith('git+')]
requirements = [r for r in requirements if not r.startswith('git+')]

if len(git_requirements) > 0:
    print("User must install \n" +
          "\n".join('{}'.format(r) for r in git_requirements) +
          "\n\nmanually")


setup(
    name='timechart',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    # Author details
    author='SLAC National Accelerator Laboratory',

    packages=find_packages(),
    package_dir={'timechart':'timechart', 'timechart_launcher':'timechart_launcher'},
    description='Time Chart Tool based on PyDM',
    url='https://github.com/slaclab/timechart',
    entry_points={
        'gui_scripts': [
            'timechart=timechart_launcher.main:main'
        ]
    },
    license='BSD',
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=requirements,
)



"""
Export the Current TimeChart Configurations into a JSON File
"""

from collections import OrderedDict
import json

from pydm import utilities
from timechart import __version__ as ver


class SettingsExporter:
    """
    Export the current configuration data into a file.
    """
    def __init__(self, pydm_main_display, include_pvs, include_chart_settings):
        """
        Parameters
        ----------
        pydm_main_display : PyDMDisplay
            The Main Window object
        include_pvs : bool
            True is to export the names and settings of the current PVs plotted in the chart;
            False if not include_chart_settings : bool
            True is to export the current overall chart settings; False if not
        """
        self.main_display = pydm_main_display
        self.include_pvs = include_pvs
        self.include_chart_settings = include_chart_settings

    def export_settings(self, filename):
        """
        Export the current settings into a JSON file.

        Parameters
        ----------
        filename : str
            The filename to export the configuration data to.
        """
        settings = OrderedDict()
        settings["__version__"] = ver
        settings["pvs"] = OrderedDict()
        settings["chart_settings"] = OrderedDict()
        chart = self.main_display.chart

        if self.include_pvs:
            pv_list = list()
            for k, v in self.main_display.channel_map.items():
                curve_settings = OrderedDict()
                curve = chart.findCurve(k)
                curve_settings["is_visible"] = curve.isVisible()
                curve_settings["color"] = v.color_string
                curve_settings["y_channel"] = v.address
                curve_settings["line_style"] = v.lineStyle
                curve_settings["line_width"] = v.lineWidth
                curve_settings["symbol"] = v.symbol
                curve_settings["symbol_size"] = v.symbolSize

                pv_list.append((k, curve_settings))
            for item in pv_list:
                settings["pvs"][item[0]] = item[1]
        if self.include_chart_settings:
            chart_settings = OrderedDict()
            chart_settings["title"] = chart.getPlotTitle()

            x_axis_label = chart.labels["bottom"]
            if x_axis_label:
                x_axis_label = x_axis_label[
                               x_axis_label.find(" -- ") + len(" -- "):]
            chart_settings["x_axis_label"] = x_axis_label
            chart_settings["x_axis_unit"] = chart.units["bottom"]

            chart_settings["left_y_axis_label"] = chart.labels["left"]
            chart_settings["left_y_axis_unit"] = chart.units["left"]

            chart_settings["show_right_y_axis"] = chart.getShowRightAxis()
            chart_settings["right_y_axis_label"] = chart.labels["right"]
            chart_settings["right_y_axis_unit"] = chart.units["right"]

            chart_settings["redraw_rate"] = chart.maxRedrawRate
            chart_settings[
                "data_sampling_mode"] = self.main_display.data_sampling_mode
            chart_settings["update_interval_hz"] = 1 / chart.getUpdateInterval()
            chart_settings[
                "limit_time_span"] = self.main_display.chart_limit_time_span_chk.isChecked()

            time_span_limit_hours = self.main_display.chart_limit_time_span_minutes_spin_box.text()
            time_span_limit_minutes = self.main_display.chart_limit_time_span_minutes_spin_box.text()
            time_span_limit_seconds = self.main_display.chart_limit_time_span_seconds_spin_box.text()

            chart_settings[
                "time_span_limit_hours"] = time_span_limit_hours if time_span_limit_hours else 0
            chart_settings[
                "time_span_limit_minutes"] = time_span_limit_minutes if time_span_limit_minutes else 0
            chart_settings[
                "time_span_limit_seconds"] = time_span_limit_seconds if time_span_limit_seconds else 0

            chart_settings["buffer_size"] = chart.getBufferSize()
            chart_settings["show_legend"] = chart.getShowLegend()
            chart_settings["background_color"] = str(
                utilities.colors.svg_color_from_hex(
                    chart.getBackgroundColor().name(), hex_on_fail=True))
            chart_settings["axis_color"] = str(
                utilities.colors.svg_color_from_hex(
                    chart.getAxisColor().name(), hex_on_fail=True))
            chart_settings["show_x_grid"] = chart.getShowXGrid()
            chart_settings["show_y_grid"] = chart.getShowYGrid()
            chart_settings["grid_alpha"] = self.main_display.gridAlpha * 10.0
            settings["chart_settings"].update(chart_settings)

        with open(filename, 'w') as json_file:
            json.dump(settings, json_file, separators=(',', ':'), indent=4)
